# Ldaptor, a Pure-Python library for LDAP
# Copyright (C) 2003 Tommi Virtanen
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of version 2.1 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""LDAP protocol server"""

from ldaptor import interfaces, delta
from ldaptor.protocols import pureldap, pureber
from ldaptor.protocols.ldap import distinguishedname, ldaperrors
from twisted.python import log
from twisted.internet import protocol, defer


class LDAPServerConnectionLostException(ldaperrors.LDAPException):
    pass


class BaseLDAPServer(protocol.Protocol):
    debug = False

    def __init__(self):
        self.buffer = b""
        self.connected = None

    berdecoder = pureldap.LDAPBERDecoderContext_TopLevel(
        inherit=pureldap.LDAPBERDecoderContext_LDAPMessage(
            fallback=pureldap.LDAPBERDecoderContext(
                fallback=pureber.BERDecoderContext()
            ),
            inherit=pureldap.LDAPBERDecoderContext(
                fallback=pureber.BERDecoderContext()
            ),
        )
    )

    def dataReceived(self, recd):
        self.buffer += recd
        while 1:
            try:
                o, bytes = pureber.berDecodeObject(self.berdecoder, self.buffer)
            except pureber.BERExceptionInsufficientData:
                o, bytes = None, 0
            self.buffer = self.buffer[bytes:]
            if o is None:
                break
            self.handle(o)

    def connectionMade(self):
        """TCP connection has opened"""
        self.connected = 1

    def connectionLost(self, reason=protocol.connectionDone):
        """Called when TCP connection has been lost"""
        self.connected = 0

    def queue(self, id, op):
        if not self.connected:
            raise LDAPServerConnectionLostException()
        msg = pureldap.LDAPMessage(op, id=id)
        if self.debug:
            log.msg("S->C %s" % repr(msg), debug=True)
        self.transport.write(msg.toWire())

    def unsolicitedNotification(self, msg):
        log.msg("Got unsolicited notification: %s" % repr(msg))

    def checkControls(self, controls):
        if controls is not None:
            for controlType, criticality, controlValue in controls:
                if criticality:
                    raise ldaperrors.LDAPUnavailableCriticalExtension(
                        b"Unknown control %s" % controlType
                    )

    def handleUnknown(self, request, controls, callback):
        log.msg("Unknown request: %r" % request)
        msg = pureldap.LDAPExtendedResponse(
            resultCode=ldaperrors.LDAPProtocolError.resultCode,
            responseName="1.3.6.1.4.1.1466.20036",
            errorMessage="Unknown request",
        )
        return msg

    def _cbLDAPError(self, reason, name):
        reason.trap(ldaperrors.LDAPException)
        return self._callErrorHandler(
            name=name,
            resultCode=reason.value.resultCode,
            errorMessage=reason.value.message,
        )

    def _cbHandle(self, response, id):
        if response is not None:
            self.queue(id, response)

    def failDefault(self, resultCode, errorMessage):
        return pureldap.LDAPExtendedResponse(
            resultCode=resultCode,
            responseName="1.3.6.1.4.1.1466.20036",
            errorMessage=errorMessage,
        )

    def _callErrorHandler(self, name, resultCode, errorMessage):
        errh = getattr(self, "fail_" + name, self.failDefault)
        return errh(resultCode=resultCode, errorMessage=errorMessage)

    def _cbOtherError(self, reason, name):
        return self._callErrorHandler(
            name=name,
            resultCode=ldaperrors.LDAPProtocolError.resultCode,
            errorMessage=reason.getErrorMessage(),
        )

    def handle(self, msg):
        assert isinstance(msg.value, pureldap.LDAPProtocolRequest)
        if self.debug:
            log.msg("S<-C %s" % repr(msg), debug=True)

        if msg.id == 0:
            self.unsolicitedNotification(msg.value)
        else:
            name = msg.value.__class__.__name__
            handler = getattr(self, "handle_" + name, self.handleUnknown)
            d = defer.maybeDeferred(
                handler,
                msg.value,
                msg.controls,
                lambda response: self._cbHandle(response, msg.id),
            )
            d.addErrback(self._cbLDAPError, name)
            d.addErrback(defer.logError)
            d.addErrback(self._cbOtherError, name)
            d.addCallback(self._cbHandle, msg.id)


class LDAPServer(BaseLDAPServer):
    """An LDAP server"""

    boundUser = None

    fail_LDAPBindRequest = pureldap.LDAPBindResponse

    def handle_LDAPBindRequest(self, request, controls, reply):
        if request.version != 3:
            raise ldaperrors.LDAPProtocolError(
                "Version %u not supported" % request.version
            )

        self.checkControls(controls)

        if request.dn == b"":
            # anonymous bind
            self.boundUser = None
            return pureldap.LDAPBindResponse(resultCode=0)
        else:
            dn = distinguishedname.DistinguishedName(request.dn)
            root = interfaces.IConnectedLDAPEntry(self.factory)
            d = root.lookup(dn)

            def _noEntry(fail):
                fail.trap(ldaperrors.LDAPNoSuchObject)
                return None

            d.addErrback(_noEntry)

            def _gotEntry(entry, auth):
                if entry is None:
                    raise ldaperrors.LDAPInvalidCredentials()

                d = entry.bind(auth)

                def _cb(entry):
                    self.boundUser = entry
                    msg = pureldap.LDAPBindResponse(
                        resultCode=ldaperrors.Success.resultCode,
                        matchedDN=entry.dn.getText(),
                    )
                    return msg

                d.addCallback(_cb)
                return d

            d.addCallback(_gotEntry, request.auth)

            return d

    def handle_LDAPUnbindRequest(self, request, controls, reply):
        # explicitly do not check unsupported critical controls -- we
        # have no way to return an error, anyway.
        self.transport.loseConnection()

    def getRootDSE(self, request, reply):
        root = interfaces.IConnectedLDAPEntry(self.factory)
        reply(
            pureldap.LDAPSearchResultEntry(
                objectName="",
                attributes=[
                    ("supportedLDAPVersion", ["3"]),
                    ("namingContexts", [root.dn.getText()]),
                    (
                        "supportedExtension",
                        [
                            pureldap.LDAPPasswordModifyRequest.oid,
                        ],
                    ),
                ],
            )
        )
        return pureldap.LDAPSearchResultDone(resultCode=ldaperrors.Success.resultCode)

    fail_LDAPCompareRequest = pureldap.LDAPCompareResponse

    def handle_LDAPCompareRequest(self, request, controls, reply):
        def _cbCompareGotBase(base, ava, reply):
            def _done(result_list):
                if result_list:
                    resultCode = ldaperrors.LDAPCompareTrue.resultCode
                else:
                    resultCode = ldaperrors.LDAPCompareFalse.resultCode
                return pureldap.LDAPCompareResponse(resultCode)

            # base.search only works with Filter Objects, and not with
            # AttributeValueAssertion objects. Here we convert the AVA to an
            # equivalent Filter so we can re-use the existing search
            # functionality we require.
            search_filter = pureldap.LDAPFilter_equalityMatch(
                attributeDesc=ava.attributeDesc, assertionValue=ava.assertionValue
            )

            d = base.search(
                filterObject=search_filter,
                scope=pureldap.LDAP_SCOPE_baseObject,
                derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
            )

            d.addCallback(_done)

            return d

        def _cbCompareLDAPError(reason):
            reason.trap(ldaperrors.LDAPException)
            return pureldap.LDAPCompareResponse(resultCode=reason.value.resultCode)

        def _cbCompareOtherError(reason):
            return pureldap.LDAPCompareResponse(
                resultCode=ldaperrors.other, errorMessage=reason.getErrorMessage()
            )

        self.checkControls(controls)
        dn = distinguishedname.DistinguishedName(request.entry)
        root = interfaces.IConnectedLDAPEntry(self.factory)

        d = root.lookup(dn)
        d.addCallback(_cbCompareGotBase, request.ava, reply)
        d.addErrback(_cbCompareLDAPError)
        d.addErrback(defer.logError)
        d.addErrback(_cbCompareOtherError)
        return d

    def _cbSearchGotBase(self, base, dn, request, reply):
        def _sendEntryToClient(entry):
            requested_attribs = request.attributes
            if len(requested_attribs) > 0 and b"*" not in requested_attribs:
                filtered_attribs = [
                    (k, entry.get(k)) for k in requested_attribs if k in entry
                ]
            else:
                filtered_attribs = entry.items()
            reply(
                pureldap.LDAPSearchResultEntry(
                    objectName=entry.dn.getText(),
                    attributes=filtered_attribs,
                )
            )

        d = base.search(
            filterObject=request.filter,
            attributes=request.attributes,
            scope=request.scope,
            derefAliases=request.derefAliases,
            sizeLimit=request.sizeLimit,
            timeLimit=request.timeLimit,
            typesOnly=request.typesOnly,
            callback=_sendEntryToClient,
        )

        def _done(_):
            return pureldap.LDAPSearchResultDone(
                resultCode=ldaperrors.Success.resultCode
            )

        d.addCallback(_done)
        return d

    def _cbSearchLDAPError(self, reason):
        reason.trap(ldaperrors.LDAPException)
        return pureldap.LDAPSearchResultDone(resultCode=reason.value.resultCode)

    def _cbSearchOtherError(self, reason):
        return pureldap.LDAPSearchResultDone(
            resultCode=ldaperrors.other, errorMessage=reason.getErrorMessage()
        )

    fail_LDAPSearchRequest = pureldap.LDAPSearchResultDone

    def handle_LDAPSearchRequest(self, request, controls, reply):
        self.checkControls(controls)

        if (
            request.baseObject == b""
            and request.scope == pureldap.LDAP_SCOPE_baseObject
            and request.filter == pureldap.LDAPFilter_present("objectClass")
        ):
            return self.getRootDSE(request, reply)
        dn = distinguishedname.DistinguishedName(request.baseObject)
        root = interfaces.IConnectedLDAPEntry(self.factory)
        d = root.lookup(dn)
        d.addCallback(self._cbSearchGotBase, dn, request, reply)
        d.addErrback(self._cbSearchLDAPError)
        d.addErrback(defer.logError)
        d.addErrback(self._cbSearchOtherError)
        return d

    fail_LDAPDelRequest = pureldap.LDAPDelResponse

    def handle_LDAPDelRequest(self, request, controls, reply):
        self.checkControls(controls)

        dn = distinguishedname.DistinguishedName(request.value)
        root = interfaces.IConnectedLDAPEntry(self.factory)
        d = root.lookup(dn)

        def _gotEntry(entry):
            d = entry.delete()
            return d

        def _report(entry):
            return pureldap.LDAPDelResponse(resultCode=0)

        d.addCallback(_gotEntry)
        d.addCallback(_report)
        return d

    fail_LDAPAddRequest = pureldap.LDAPAddResponse

    def handle_LDAPAddRequest(self, request, controls, reply):
        self.checkControls(controls)

        attributes = {}
        for name, vals in request.attributes:
            attributes.setdefault(name.value, set())
            attributes[name.value].update([x.value for x in vals])
        dn = distinguishedname.DistinguishedName(request.entry)
        rdn = dn.split()[0].getText()
        parent = dn.up()
        root = interfaces.IConnectedLDAPEntry(self.factory)
        d = root.lookup(parent)

        def _gotEntry(parent):
            d = parent.addChild(rdn, attributes)
            return d

        def _report(entry):
            return pureldap.LDAPAddResponse(resultCode=0)

        d.addCallback(_gotEntry)
        d.addCallback(_report)
        return d

    fail_LDAPModifyDNRequest = pureldap.LDAPModifyDNResponse

    def handle_LDAPModifyDNRequest(self, request, controls, reply):
        self.checkControls(controls)
        dn = distinguishedname.DistinguishedName(request.entry)
        newrdn = distinguishedname.RelativeDistinguishedName(request.newrdn)
        deleteoldrdn = bool(request.deleteoldrdn)
        if not deleteoldrdn:
            raise ldaperrors.LDAPUnwillingToPerform(
                "Cannot handle preserving old RDN yet."
            )
        newSuperior = request.newSuperior
        if newSuperior is None:
            newSuperior = dn.up()
        else:
            newSuperior = distinguishedname.DistinguishedName(newSuperior)
        newdn = distinguishedname.DistinguishedName(
            listOfRDNs=(newrdn,) + newSuperior.split()
        )
        root = interfaces.IConnectedLDAPEntry(self.factory)
        d = root.lookup(dn)

        def _gotEntry(entry):
            d = entry.move(newdn)
            return d

        def _report(entry):
            return pureldap.LDAPModifyDNResponse(resultCode=0)

        d.addCallback(_gotEntry)
        d.addCallback(_report)
        return d

    fail_LDAPModifyRequest = pureldap.LDAPModifyResponse

    def handle_LDAPModifyRequest(self, request, controls, reply):
        self.checkControls(controls)

        root = interfaces.IConnectedLDAPEntry(self.factory)
        mod = delta.ModifyOp.fromLDAP(request)
        d = mod.patch(root)

        def _patched(entry):
            return entry.commit()

        def _report(entry):
            return pureldap.LDAPModifyResponse(resultCode=0)

        d.addCallback(_patched)
        d.addCallback(_report)
        return d

    fail_LDAPExtendedRequest = pureldap.LDAPExtendedResponse

    def handle_LDAPExtendedRequest(self, request, controls, reply):
        self.checkControls(controls)

        for handler in [
            getattr(self, attr)
            for attr in dir(self)
            if attr.startswith("extendedRequest_")
        ]:
            if getattr(handler, "oid", None) == request.requestName:
                berdecoder = getattr(handler, "berdecoder", None)

                if berdecoder is None:
                    values = [request.requestValue]
                else:
                    values = pureber.berDecodeMultiple(request.requestValue, berdecoder)

                d = defer.maybeDeferred(handler, *values, **{"reply": reply})

                def eb(fail, oid):
                    fail.trap(ldaperrors.LDAPException)
                    return pureldap.LDAPExtendedResponse(
                        resultCode=fail.value.resultCode,
                        errorMessage=fail.value.message,
                        responseName=oid,
                    )

                d.addErrback(eb, request.requestName)
                return d

        raise ldaperrors.LDAPProtocolError(
            b"Unknown extended request: %s" % request.requestName
        )

    def extendedRequest_LDAPPasswordModifyRequest(self, data, reply):
        if not isinstance(data, pureber.BERSequence):
            raise ldaperrors.LDAPProtocolError(
                "Extended request PasswordModify expected a BERSequence."
            )

        userIdentity = None
        oldPasswd = None
        newPasswd = None

        for value in data:
            if isinstance(value, pureldap.LDAPPasswordModifyRequest_userIdentity):
                if userIdentity is not None:
                    raise ldaperrors.LDAPProtocolError(
                        "Extended request "
                        "PasswordModify received userIdentity twice."
                    )
                userIdentity = value.value
            elif isinstance(value, pureldap.LDAPPasswordModifyRequest_oldPasswd):
                if oldPasswd is not None:
                    raise ldaperrors.LDAPProtocolError(
                        "Extended request PasswordModify " "received oldPasswd twice."
                    )
                oldPasswd = value.value
            elif isinstance(value, pureldap.LDAPPasswordModifyRequest_newPasswd):
                if newPasswd is not None:
                    raise ldaperrors.LDAPProtocolError(
                        "Extended request PasswordModify " "received newPasswd twice."
                    )
                newPasswd = value.value
            else:
                raise ldaperrors.LDAPProtocolError(
                    "Extended request PasswordModify " "received unexpected item."
                )

        if self.boundUser is None:
            raise ldaperrors.LDAPStrongAuthRequired()

        if userIdentity is not None and userIdentity != self.boundUser.dn:
            log.msg(
                "User {actor} tried to change password of {target}".format(
                    actor=self.boundUser.dn.getText(),
                    target=userIdentity,
                )
            )
            raise ldaperrors.LDAPInsufficientAccessRights()
        if oldPasswd is not None or newPasswd is None:
            raise ldaperrors.LDAPOperationsError("Password does not support this case.")
        self.boundUser.setPassword(newPasswd)
        d = self.boundUser.commit()

        def cb_(result):
            if result:
                return pureldap.LDAPExtendedResponse(
                    resultCode=ldaperrors.Success.resultCode,
                    responseName=self.extendedRequest_LDAPPasswordModifyRequest.oid,
                )
            else:
                raise ldaperrors.LDAPOperationsError("Internal error.")

        d.addCallback(cb_)
        return d

    extendedRequest_LDAPPasswordModifyRequest.oid = (
        pureldap.LDAPPasswordModifyRequest.oid
    )
    extendedRequest_LDAPPasswordModifyRequest.berdecoder = pureber.BERDecoderContext(
        inherit=pureldap.LDAPBERDecoderContext_LDAPPasswordModifyRequest(
            inherit=pureber.BERDecoderContext()
        )
    )


if __name__ == "__main__":
    """
    Demonstration LDAP server; reads LDIF from stdin and
    serves that over LDAP on port 10389.
    """
    from twisted.internet import reactor
    import sys

    log.startLogging(sys.stderr)

    from twisted.python import components
    from ldaptor import inmemory

    class LDAPServerFactory(protocol.ServerFactory):
        def __init__(self, root):
            self.root = root

    components.registerAdapter(
        lambda x: x.root, LDAPServerFactory, interfaces.IConnectedLDAPEntry
    )

    def start(db):
        factory = LDAPServerFactory(db)
        factory.protocol = LDAPServer
        reactor.listenTCP(10389, factory)

    d = inmemory.fromLDIFFile(sys.stdin)
    d.addCallback(start)
    d.addErrback(log.err)
    reactor.run()



"""
Command line argument/options available to various ldaptor tools.
"""
from twisted.python import usage, reflect
from twisted.python.usage import UsageError
from ldaptor.protocols import pureldap
from ldaptor.protocols.ldap import distinguishedname

__all__ = [
    "Options",
    "Options_base",
    "Options_base_optional",
    "Options_bind",
    "Options_bind_mandatory",
    "Options_scope",
    "Options_service_location",
    "UsageError",
]


class Options(usage.Options):
    optParameters = ()

    def postOptions(self):
        postOpt = {}
        reflect.addMethodNamesToDict(self.__class__, postOpt, "postOptions_")
        for name in postOpt.keys():
            method = getattr(self, "postOptions_" + name)
            method()


class Options_service_location:
    """
    Mixing for providing the --service-location option.
    """

    def opt_service_location(self, value):
        """Service location, in the form BASEDN:HOST[:PORT]"""

        if "service-location" not in self.opts:
            self.opts["service-location"] = {}

        if ":" not in value:
            raise usage.UsageError("service-location must specify host")

        base, location = value.split(":", 1)
        try:
            dn = distinguishedname.DistinguishedName(base)
        except distinguishedname.InvalidRelativeDistinguishedName as e:
            raise usage.UsageError(str(e))

        if ":" in location:
            host, port = location.split(":", 1)
        else:
            host, port = location, None

        self.opts["service-location"][dn] = (host, port)

    def postOptions_service_location(self):
        if "service-location" not in self.opts:
            self.opts["service-location"] = {}


class Options_base_optional:
    optParameters = (("base", None, None, "LDAP base dn"),)


class Options_base(Options_base_optional):
    def postOptions_base(self):
        # check that some things are given
        if self.opts["base"] is None:
            raise usage.UsageError("base must be given")


class Options_scope:
    optParameters = (
        ("scope", None, "sub", "LDAP search scope (one of base, one, sub)"),
    )

    def postOptions_scope(self):
        synonyms = {
            "base": "baseObject",
            "single": "singleLevel",
            "subtree": "wholeSubtree",
            "sub": "wholeSubtree",
        }
        scope = self.opts["scope"]
        scope = synonyms.get(scope, scope)
        try:
            scope = getattr(pureldap, "LDAP_SCOPE_" + scope)
        except AttributeError:
            raise usage.UsageError("bad scope: {}".format(scope))
        self.opts["scope"] = scope


class Options_bind:
    optParameters = (
        ("binddn", None, None, "use Distinguished Name to bind to the directory"),
        ("bind-auth-fd", None, None, "read bind password from filedescriptor"),
    )

    def postOptions_bind_auth_fd_numeric(self):
        val = self.opts["bind-auth-fd"]
        if val is not None:
            try:
                val = int(val)
            except ValueError:
                raise usage.UsageError("bind-auth-fd value must be numeric")
            self.opts["bind-auth-fd"] = val


class Options_bind_mandatory(Options_bind):
    def postOptions_bind_mandatory(self):
        if not self.opts["binddn"]:
            raise usage.UsageError("binddn must be given")



"""Utilities for writing Twistedy unit tests and debugging."""

from twisted.internet import defer
from twisted.python import failure
from twisted.trial import unittest
from twisted.test import proto_helpers
from ldaptor import config
from ldaptor._encoder import to_bytes


def mustRaise(dummy):
    raise unittest.FailTest("Should have raised an exception.")


def calltrace():
    """Print out all function calls. For debug use only."""

    def printfuncnames(frame, event, arg):
        print(
            "|%s: %s:%d:%s"
            % (
                event,
                frame.f_code.co_filename,
                frame.f_code.co_firstlineno,
                frame.f_code.co_name,
            )
        )

    import sys

    sys.setprofile(printfuncnames)


class FakeTransport:
    def __init__(self, proto):
        self.proto = proto

    def loseConnection(self):
        self.proto.connectionLost()


class LDAPClientTestDriver:
    """

    A test driver that looks somewhat like a real LDAPClient.

    Pass in a list of lists of LDAPProtocolResponses. For each sent
    LDAP message, the first item of said list is iterated through, and
    all the items are sent as responses to the callback. The sent LDAP
    messages are stored in self.sent, so you can assert that the sent
    messages are what they are supposed to be.

    It is also possible to include a Failure instance instead of a list
    of LDAPProtocolResponses which will cause the errback to be called
    with the failure.
    """

    fakeUnbindResponse = "fake-unbind-by-LDAPClientTestDriver"

    def __init__(self, *responses):
        self.sent = []
        self.responses = list(responses)
        self.connected = None
        self.transport = FakeTransport(self)

    def send(self, op):
        self.sent.append(op)
        resps = self._response()
        assert len(resps) == 1, "got %d responses for a .send()" % len(resps)
        r = resps[0]
        if isinstance(r, failure.Failure):
            return defer.fail(r)
        else:
            return defer.succeed(r)

    def send_multiResponse_(
        self, op, controls, return_controls, handler, *args, **kwargs
    ):
        d = defer.Deferred()
        self.sent.append(op)
        responses = self._response()
        response_controls = None
        while responses:
            r = responses.pop(0)
            if isinstance(r, failure.Failure):
                d.errback(r)
                break
            if return_controls:
                ret = handler(r, response_controls, *args, **kwargs)
            else:
                ret = handler(r, *args, **kwargs)
            if responses:
                msg = (
                    "got %d responses still to give, "
                    "but handler wants none (got %r)."
                ) % (len(responses), ret)
                assert not ret, msg
            else:
                msg = (
                    "no more responses to give, but handler "
                    "still wants more (got %r)." % ret
                )
                assert ret, msg
        return d

    def send_multiResponse(self, op, handler, *args, **kwargs):
        return self.send_multiResponse_(op, None, False, handler, *args, **kwargs)

    def send_multiResponse_ex(self, op, controls, handler, *args, **kwargs):
        return self.send_multiResponse_(op, controls, True, handler, *args, **kwargs)

    def send_noResponse(self, op):
        if len(self.responses) == 0:
            msg = "Ran out of responses"
            assert op == self.fakeUnbindResponse, msg
        else:
            self.responses.pop(0)
        self.sent.append(op)

    def _response(self):
        assert self.responses, "Ran out of responses"
        responses = self.responses.pop(0)
        return responses

    def assertNothingSent(self):
        # just a bit more explicit
        self.assertSent()

    def assertSent(self, *shouldBeSent):
        shouldBeSent = list(shouldBeSent)
        msg = "{} expected to send {!r} but sent {!r}".format(
            self.__class__.__name__, shouldBeSent, self.sent
        )
        assert self.sent == shouldBeSent, msg
        sentStr = b"".join([to_bytes(x) for x in self.sent])
        shouldBeSentStr = b"".join([to_bytes(x) for x in shouldBeSent])
        msg = "{} expected to send data {!r} but sent {!r}".format(
            self.__class__.__name__, shouldBeSentStr, sentStr
        )
        assert sentStr == shouldBeSentStr, msg

    def connectionMade(self):
        """TCP connection has opened"""
        self.connected = 1

    def connectionLost(self, reason=None):
        """
        Called when TCP connection has been lost
        """
        msg = (
            "connectionLost called even when have "
            "responses left: %r" % self.responses
        )
        assert not self.responses, msg
        self.connected = 0

    def unbind(self):
        assert self.connected
        r = self.fakeUnbindResponse
        self.send_noResponse(r)
        self.transport.loseConnection()


def createServer(proto, *responses, **kw):
    """
    Create an LDAP server for testing.
    :param proto: The server protocol factory (e.g. `ProxyBase`).
    :param responses: The responses to initialize the `LDAPClientTestDrive`.
    :param proto_args: Optional mapping passed as keyword args to protocol factory.
    """
    if "proto_args" in kw:
        proto_args = kw["proto_args"]
        del kw["proto_args"]
    else:
        proto_args = {}

    def createClient(factory):
        factory.doStart()
        proto = factory.buildProtocol(addr=None)
        proto.connectionMade()

    overrides = kw.setdefault("serviceLocationOverrides", {})
    overrides.setdefault("", createClient)
    conf = config.LDAPConfig(**kw)
    server = proto(conf, **proto_args)
    clientTestDriver = LDAPClientTestDriver(*responses)
    server.protocol = lambda: clientTestDriver
    server.clientTestDriver = clientTestDriver
    server.transport = proto_helpers.StringTransport()
    server.connectionMade()
    return server



from copy import deepcopy

from ldaptor._encoder import get_strings


class LDAPAttributeSet(set):
    def __init__(self, key, *a, **kw):
        """
        Represents all the values for an attribute in an LDAP entry. An entry
        might have "cn" or "objectClass" or "uid" attributes, and this class
        represents each of those.

        You can find the name of the LDAP entry attribute (eg. "uid") with the
        ``.key`` member variable.

        You can find the values of the LDAP attribute by casting this to a
        ``list``.
        @param key: the key of the attribute, eg "uid".
        @type key: str
        @param args: set of values for this attribute, eg. "jsmith"
        """
        self.key = key
        super().__init__(*a, **kw)

    def __repr__(self):
        values = list(self)
        values.sort()
        attributes = ", ".join([repr(x) for x in values])
        return "{}({!r}, [{}])".format(self.__class__.__name__, self.key, attributes)

    def __eq__(self, other):
        """
        Note that LDAPAttributeSets can also be compared against any
        iterator. In that case the attributeType will be ignored.
        """
        if isinstance(other, LDAPAttributeSet):
            if self.key != other.key:
                return False
            return super().__eq__(other)
        else:
            me = list(self)
            me.sort()
            him = list(other)
            him.sort()
            return me == him

    def __ne__(self, other):
        return not self == other

    def add(self, key):
        """
        Adding key to the attributes with checking
        if it exists as byte or unicode string
        """
        for k in get_strings(key):
            if k in self:
                return

        set.add(self, key)

    def remove(self, key):
        """
        Removing key from the attributes with checking
        if it exists as byte or unicode string
        """
        for k in get_strings(key):
            if k in self:
                set.remove(self, k)
                return

        raise KeyError(key)

    def copy(self):
        result = self.__class__(self.key)
        result.update(self)
        return result

    __copy__ = copy

    def __deepcopy__(self, memo):
        result = self.__class__(self.key)
        memo[id(self)] = result
        data = deepcopy(set(self), memo)
        result.update(data)
        return result



COUNTRY = (
    "dc=fr",
    {
        "objectClass": ["dcObject", "country"],
        "dc": ["fr"],
        "description": ["French country 2 letters iso description"],
    },
)
COMPANY = (
    "dc=example",
    {
        "objectClass": ["dcObject", "organization"],
        "dc": ["example"],
        "description": ["My organisation"],
        "o": ["Example, Inc"],
    },
)
PEOPLE = (
    "ou=people",
    {
        "ou": ["people"],
        "description": ["People from Example Inc"],
        "objectclass": ["organizationalunit"],
    },
)
USERS = [
    (
        "uid=yoen",
        {
            "objectClass": ["people", "inetOrgPerson"],
            "cn": ["Yoen Van der Weld"],
            "sn": ["Van der Weld"],
            "givenName": ["Yoen"],
            "uid": ["yoen"],
            "mail": ["/home/yoen/mailDir"],
            "userPassword": ["secret"],
        },
    ),
    (
        "uid=esteban",
        {
            "objectClass": ["people", "inetOrgPerson"],
            "cn": ["Esteban Garcia Marquez"],
            "sn": ["Garcia Marquez"],
            "givenName": ["Esteban"],
            "uid": ["esteban"],
            "mail": ["/home/esteban/mailDir"],
            "userPassword": ["secret2"],
        },
    ),
    (
        "uid=mohamed",
        {
            "objectClass": ["people", "inetOrgPerson"],
            "cn": ["Mohamed Al Ghâlib"],
            "sn": ["Al Ghâlib"],
            "givenName": ["mohamed"],
            "uid": ["mohamed"],
            "mail": ["/home/mohamed/mailDir"],
            "userPassword": ["secret3"],
        },
    ),
]



"""
Test cases for LDIF directory tree writing/reading.
"""

import os
import sys
import random
import errno
import shutil
import unittest as stdlib_unittest

from twisted.python import log
from twisted.trial import unittest
from twisted.internet import testing

from ldaptor import ldiftree, entry, delta, testutil
from ldaptor.entry import BaseLDAPEntry
from ldaptor.protocols.ldap import ldaperrors, ldifprotocol
from ldaptor.test import util


def writeFile(path, content):
    f = open(path, "wb")
    f.write(content)
    f.close()


skipIfWindowsOrRoot = stdlib_unittest.skipIf(
    sys.platform == "win32" or os.getuid() == 0,
    "Can't test on windows or as root",
)


skipIfWindows = stdlib_unittest.skipIf(
    sys.platform == "win32",
    "TODO: fails on windows",
)


class RandomizeListdirTestCase(unittest.TestCase):
    """
    It patches the os.listdir to return the members in a random order.
    """

    def randomListdir(self, *args, **kwargs):
        """
        This is used to replace the os.listdir.
        """
        r = self.__os_listdir(*args, **kwargs)
        random.shuffle(r)
        return r

    def setUp(self):
        self.__os_listdir = os.listdir
        os.listdir = self.randomListdir

        def reverse_listdir():
            os.listdir = self.__os_listdir

        self.addCleanup(reverse_listdir)

    def chmod(self, path, mode):
        self.addCleanup(os.chmod, path, os.stat(path).st_mode)
        os.chmod(path, mode)


class Dir2LDIF(RandomizeListdirTestCase):
    def setUp(self):
        super().setUp()
        self.tree = self.mktemp()
        os.mkdir(self.tree)
        com = os.path.join(self.tree, "dc=com.dir")
        os.mkdir(com)
        example = os.path.join(com, "dc=example.dir")
        os.mkdir(example)
        writeFile(
            os.path.join(example, "cn=foo.ldif"),
            b"""\
dn: cn=foo,dc=example,dc=com
cn: foo
objectClass: top

""",
        )
        writeFile(
            os.path.join(example, "cn=bad-two-entries.ldif"),
            b"""\
dn: cn=bad-two-entries,dc=example,dc=com
cn: bad-two-entries
objectClass: top

dn: cn=more,dc=example,dc=com
cn: more
objectClass: top

""",
        )
        writeFile(
            os.path.join(example, "cn=bad-missing-end.ldif"),
            b"""\
dn: cn=bad-missing-end,dc=example,dc=com
cn: bad-missing-end
objectClass: top
""",
        )
        writeFile(os.path.join(example, "cn=bad-empty.ldif"), b"")
        writeFile(os.path.join(example, "cn=bad-only-newline.ldif"), b"\n")
        sales = os.path.join(example, "ou=Sales.dir")
        os.mkdir(sales)
        writeFile(
            os.path.join(sales, "cn=sales-thingie.ldif"),
            b"""\
dn: cn=sales-thingie,ou=Sales,dc=example,dc=com
cn: sales-thingie
objectClass: top

""",
        )

    def testSimpleRead(self):
        want = BaseLDAPEntry(
            dn=b"cn=foo,dc=example,dc=com",
            attributes={
                b"objectClass": [b"top"],
                b"cn": [b"foo"],
            },
        )
        d = ldiftree.get(self.tree, want.dn)
        d.addCallback(self.failUnlessEqual, want)
        return d

    @skipIfWindowsOrRoot
    def testNoAccess(self):
        self.chmod(
            os.path.join(self.tree, "dc=com.dir", "dc=example.dir", "cn=foo.ldif"), 0
        )
        d = ldiftree.get(self.tree, "cn=foo,dc=example,dc=com")

        def eb(fail):
            fail.trap(IOError)
            self.assertEqual(fail.value.errno, errno.EACCES)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def gettingDNRaises(self, dn, exceptionClass):
        d = ldiftree.get(self.tree, dn)

        def eb(fail):
            fail.trap(exceptionClass)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def testMultipleError(self):
        return self.gettingDNRaises(
            "cn=bad-two-entries,dc=example,dc=com",
            ldiftree.LDIFTreeEntryContainsMultipleEntries,
        )

    def testMissingEndError(self):
        return self.gettingDNRaises(
            "cn=bad-missing-end,dc=example,dc=com",
            ldiftree.LDIFTreeEntryContainsNoEntries,
        )

    def testEmptyError(self):
        return self.gettingDNRaises(
            "cn=bad-empty,dc=example,dc=com", ldiftree.LDIFTreeEntryContainsNoEntries
        )

    def testOnlyNewlineError(self):
        return self.gettingDNRaises(
            "cn=bad-only-newline,dc=example,dc=com",
            ldifprotocol.LDIFLineWithoutSemicolonError,
        )

    def testTreeBranches(self):
        want = BaseLDAPEntry(
            dn=b"cn=sales-thingie,ou=Sales,dc=example,dc=com",
            attributes={
                b"objectClass": [b"top"],
                b"cn": [b"sales-thingie"],
            },
        )
        d = ldiftree.get(self.tree, want.dn)
        d.addCallback(self.failUnlessEqual, want)
        return d


class LDIF2Dir(RandomizeListdirTestCase):
    def setUp(self):
        super().setUp()
        self.tree = self.mktemp()
        os.mkdir(self.tree)
        com = os.path.join(self.tree, "dc=com.dir")
        os.mkdir(com)
        example = os.path.join(com, "dc=example.dir")
        os.mkdir(example)
        writeFile(
            os.path.join(example, "cn=pre-existing.ldif"),
            b"""\
dn: cn=pre-existing,dc=example,dc=com
cn: pre-existing
objectClass: top

""",
        )
        writeFile(
            os.path.join(example, "ou=OrgUnit.ldif"),
            b"""\
dn: ou=OrgUnit,dc=example,dc=com
ou: OrgUnit
objectClass: organizationalUnit

""",
        )

    def testSimpleWrite(self):
        e = BaseLDAPEntry(
            dn="cn=foo,dc=example,dc=com",
            attributes={
                "objectClass": ["top"],
                "cn": ["foo"],
            },
        )
        d = ldiftree.put(self.tree, e)
        d.addCallback(self._cb_testSimpleWrite)
        return d

    def _cb_testSimpleWrite(self, entry):
        path = os.path.join(self.tree, "dc=com.dir", "dc=example.dir", "cn=foo.ldif")
        self.failUnless(os.path.isfile(path))
        self.failUnlessEqual(
            open(path, "rb").read(),
            b"""\
dn: cn=foo,dc=example,dc=com
objectClass: top
cn: foo

""",
        )

    def testDirCreation(self):
        e = BaseLDAPEntry(
            dn="cn=create-me,ou=OrgUnit,dc=example,dc=com",
            attributes={
                "objectClass": ["top"],
                "cn": ["create-me"],
            },
        )
        d = ldiftree.put(self.tree, e)
        d.addCallback(self._cb_testDirCreation)
        return d

    def _cb_testDirCreation(self, entry):
        path = os.path.join(
            self.tree,
            "dc=com.dir",
            "dc=example.dir",
            "ou=OrgUnit.dir",
            "cn=create-me.ldif",
        )
        self.failUnless(os.path.isfile(path))
        self.failUnlessEqual(
            open(path, "rb").read(),
            b"""\
dn: cn=create-me,ou=OrgUnit,dc=example,dc=com
objectClass: top
cn: create-me

""",
        )

    def testDirExists(self):
        e = BaseLDAPEntry(
            dn="cn=create-me,ou=OrgUnit,dc=example,dc=com",
            attributes={
                "objectClass": ["top"],
                "cn": ["create-me"],
            },
        )
        dirpath = os.path.join(
            self.tree, "dc=com.dir", "dc=example.dir", "ou=OrgUnit.dir"
        )
        os.mkdir(dirpath)
        d = ldiftree.put(self.tree, e)
        d.addCallback(self._cb_testDirExists, dirpath)
        return d

    def _cb_testDirExists(self, entry, dirpath):
        path = os.path.join(dirpath, "cn=create-me.ldif")
        self.failUnless(os.path.isfile(path))
        self.failUnlessEqual(
            open(path, "rb").read(),
            b"""\
dn: cn=create-me,ou=OrgUnit,dc=example,dc=com
objectClass: top
cn: create-me

""",
        )

    def testMissingLinkError(self):
        e = BaseLDAPEntry(
            dn="cn=bad-create,ou=NoSuchOrgUnit,dc=example,dc=com",
            attributes={
                "objectClass": ["top"],
                "cn": ["bad-create"],
            },
        )

        d = ldiftree.put(self.tree, e)

        failure = self.failureResultOf(d)
        self.assertIsInstance(failure.value, ldiftree.LDIFTreeNoSuchObject)

    def testAddTopLevel(self):
        e = BaseLDAPEntry(
            dn="dc=org",
            attributes={
                "objectClass": ["dcObject"],
                "dc": ["org"],
            },
        )
        d = ldiftree.put(self.tree, e)
        d.addCallback(self._cb_testAddTopLevel)
        return d

    def _cb_testAddTopLevel(self, entry):
        path = os.path.join(self.tree, "dc=org.ldif")
        self.failUnless(os.path.isfile(path))
        self.failUnlessEqual(
            open(path, "rb").read(),
            b"""\
dn: dc=org
objectClass: dcObject
dc: org

""",
        )


class LDIFTreeEntryTests(RandomizeListdirTestCase):
    """
    Tests for LDIFTreeEntry.
    """

    # TODO share the actual tests with inmemory and any other
    # implementations of the same interface
    def setUp(self):
        super().setUp()
        self.tree = self.mktemp()
        os.mkdir(self.tree)
        com = os.path.join(self.tree, "dc=com.dir")
        os.mkdir(com)
        example = os.path.join(com, "dc=example.dir")
        os.mkdir(example)
        meta = os.path.join(example, "ou=metasyntactic.dir")
        os.mkdir(meta)
        writeFile(
            os.path.join(example, "ou=metasyntactic.ldif"),
            b"""\
dn: ou=metasyntactic,dc=example,dc=com
objectClass: a
objectClass: b
ou: metasyntactic

""",
        )
        foo = os.path.join(meta, "cn=foo.dir")
        writeFile(
            os.path.join(meta, "cn=foo.ldif"),
            b"""\
dn: cn=foo,ou=metasyntactic,dc=example,dc=com
objectClass: a
objectClass: b
cn: foo

""",
        )
        bar = os.path.join(meta, "cn=bar.dir")
        writeFile(
            os.path.join(meta, "cn=bar.ldif"),
            b"""\
dn: cn=bar,ou=metasyntactic,dc=example,dc=com
objectClass: a
objectClass: b
cn: bar

""",
        )
        empty = os.path.join(example, "ou=empty.dir")
        writeFile(
            os.path.join(example, "ou=empty.ldif"),
            b"""\
dn: ou=empty,dc=example,dc=com
objectClass: a
objectClass: b
ou: empty

""",
        )
        oneChild = os.path.join(example, "ou=oneChild.dir")
        os.mkdir(oneChild)
        writeFile(
            os.path.join(example, "ou=oneChild.ldif"),
            b"""\
dn: ou=oneChild,dc=example,dc=com
objectClass: a
objectClass: b
ou: oneChild

""",
        )
        theChild = os.path.join(oneChild, "cn=theChild.dir")
        writeFile(
            os.path.join(oneChild, "cn=theChild.ldif"),
            b"""\
dn: cn=theChild,ou=oneChild,dc=example,dc=com
objectClass: a
objectClass: b
cn: theChild

""",
        )
        # Invalid file
        writeFile(os.path.join(oneChild, "cn=invalidChild.lddd"), b"invalid data")

        self.root = ldiftree.LDIFTreeEntry(self.tree)
        self.example = ldiftree.LDIFTreeEntry(example, "dc=example,dc=com")
        self.empty = ldiftree.LDIFTreeEntry(empty, "ou=empty,dc=example,dc=com")
        self.meta = ldiftree.LDIFTreeEntry(meta, "ou=metasyntactic,dc=example,dc=com")
        self.foo = ldiftree.LDIFTreeEntry(
            foo, "cn=foo,ou=metasyntactic,dc=example,dc=com"
        )
        self.bar = ldiftree.LDIFTreeEntry(
            bar, "cn=bar,ou=metasyntactic,dc=example,dc=com"
        )
        self.oneChild = ldiftree.LDIFTreeEntry(
            oneChild, "ou=oneChild,dc=example,dc=com"
        )
        self.theChild = ldiftree.LDIFTreeEntry(
            theChild, "cn=theChild,ou=oneChild,dc=example,dc=com"
        )

    def test_children_empty(self):
        d = self.empty.children()

        def cb(children):
            self.assertEqual(children, [])

        d.addCallback(cb)
        return d

    def test_children_oneChild(self):
        d = self.oneChild.children()
        d.addCallback(self._cb_test_children_oneChild)
        return d

    def _cb_test_children_oneChild(self, children):
        self.assertEqual(len(children), 1)
        got = [e.dn for e in children]
        want = ["cn=theChild,ou=oneChild,dc=example,dc=com"]
        got.sort()
        want.sort()
        self.assertEqual(got, want)

    def test_children_repeat(self):
        """Test that .children() returns a copy of the data so that modifying it does not affect behaviour."""
        d = self.oneChild.children()
        d.addCallback(self._cb_test_children_repeat_1)
        return d

    def _cb_test_children_repeat_1(self, children1):
        self.assertEqual(len(children1), 1)

        children1.pop()

        d = self.oneChild.children()
        d.addCallback(self._cb_test_children_repeat_2)
        return d

    def _cb_test_children_repeat_2(self, children2):
        self.assertEqual(len(children2), 1)

    def test_children_twoChildren(self):
        d = self.meta.children()
        d.addCallback(self._cb_test_children_twoChildren)
        return d

    def _cb_test_children_twoChildren(self, children):
        self.assertEqual(len(children), 2)
        want = [
            "cn=foo,ou=metasyntactic,dc=example,dc=com",
            "cn=bar,ou=metasyntactic,dc=example,dc=com",
        ]
        got = [e.dn for e in children]
        got.sort()
        want.sort()
        self.assertEqual(got, want)

    def test_children_twoChildren_callback(self):
        children = []
        d = self.meta.children(callback=children.append)
        d.addCallback(self._cb_test_children_twoChildren_callback, children)
        return d

    def _cb_test_children_twoChildren_callback(self, r, children):
        self.assertIdentical(r, None)
        self.assertEqual(len(children), 2)
        want = [
            "cn=foo,ou=metasyntactic,dc=example,dc=com",
            "cn=bar,ou=metasyntactic,dc=example,dc=com",
        ]
        got = [e.dn for e in children]
        got.sort()
        want.sort()
        self.assertEqual(got, want)

    @skipIfWindowsOrRoot
    def test_children_noAccess_dir_noRead(self):
        self.chmod(self.meta.path, 0o300)
        d = self.meta.children()

        def eb(fail):
            fail.trap(OSError)
            self.assertEqual(fail.value.errno, errno.EACCES)
            self.chmod(self.meta.path, 0o755)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    @skipIfWindowsOrRoot
    def test_children_noAccess_dir_noExec(self):
        self.chmod(self.meta.path, 0o600)
        d = self.meta.children()

        def eb(fail):
            fail.trap(IOError)
            self.assertEqual(fail.value.errno, errno.EACCES)
            self.chmod(self.meta.path, 0o755)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    @skipIfWindowsOrRoot
    def test_children_noAccess_file(self):
        self.chmod(os.path.join(self.meta.path, "cn=foo.ldif"), 0)
        d = self.meta.children()

        def eb(fail):
            fail.trap(IOError)
            self.assertEqual(fail.value.errno, errno.EACCES)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_addChild(self):
        self.empty.addChild(
            rdn="a=b",
            attributes={
                "objectClass": ["a", "b"],
                "a": "b",
            },
        )
        d = self.empty.children()
        d.addCallback(self._cb_test_addChild)
        return d

    def _cb_test_addChild(self, children):
        self.assertEqual(len(children), 1)
        got = [e.dn for e in children]
        want = [
            "a=b,ou=empty,dc=example,dc=com",
        ]
        got.sort()
        want.sort()
        self.assertEqual(got, want)

    def test_addChild_Exists(self):
        self.assertRaises(
            ldaperrors.LDAPEntryAlreadyExists,
            self.meta.addChild,
            rdn="cn=foo",
            attributes={
                "objectClass": ["a"],
                "cn": "foo",
            },
        )

    def test_parent(self):
        self.assertEqual(self.foo.parent(), self.meta)
        self.assertEqual(self.meta.parent(), self.example)
        self.assertEqual(self.root.parent(), None)

    def test_subtree_empty(self):
        d = self.empty.subtree()
        d.addCallback(self._cb_test_subtree_empty)
        return d

    def _cb_test_subtree_empty(self, entries):
        self.assertEqual(len(entries), 1)

    def test_subtree_oneChild(self):
        d = self.oneChild.subtree()
        d.addCallback(self._cb_test_subtree_oneChild)
        return d

    def _cb_test_subtree_oneChild(self, results):
        got = results
        want = [
            self.oneChild,
            self.theChild,
        ]
        self.assertEqual(got, want)

    def test_subtree_oneChild_cb(self):
        got = []
        d = self.oneChild.subtree(got.append)
        d.addCallback(self._cb_test_subtree_oneChild_cb, got)
        return d

    def _cb_test_subtree_oneChild_cb(self, r, got):
        self.assertEqual(r, None)

        want = [
            self.oneChild,
            self.theChild,
        ]
        self.assertEqual(got, want)

    def test_subtree_many(self):
        deferred = self.example.subtree()

        result = self.successResultOf(deferred)

        expected = [
            self.example,
            self.oneChild,
            self.theChild,
            self.empty,
            self.meta,
            self.bar,
            self.foo,
        ]
        self.assertCountEqual(expected, result)

    def test_subtree_many_cb(self):
        got = []
        deferred = self.example.subtree(callback=got.append)

        result = self.successResultOf(deferred)

        self.assertIsNone(result)
        expected = [
            self.example,
            self.oneChild,
            self.theChild,
            self.empty,
            self.meta,
            self.bar,
            self.foo,
        ]
        self.assertCountEqual(expected, got)

    def test_lookup_fail(self):
        dn = "cn=thud,ou=metasyntactic,dc=example,dc=com"
        d = self.root.lookup(dn)

        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)
            self.assertEqual(fail.value.message, dn)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_lookup_fail_outOfTree(self):
        dn = "dc=invalid"
        d = self.root.lookup(dn)

        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)
            self.assertEqual(fail.value.message, dn)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_lookup_fail_outOfTree_2(self):
        dn = "dc=invalid"
        d = self.example.lookup(dn)

        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)
            self.assertEqual(fail.value.message, dn)

        d.addCallbacks(testutil.mustRaise, eb)

    def test_lookup_fail_multipleError(self):
        writeFile(
            os.path.join(self.example.path, "cn=bad-two-entries.ldif"),
            b"""\
dn: cn=bad-two-entries,dc=example,dc=com
cn: bad-two-entries
objectClass: top

dn: cn=more,dc=example,dc=com
cn: more
objectClass: top

""",
        )
        self.assertRaises(
            ldiftree.LDIFTreeEntryContainsMultipleEntries,
            self.example.lookup,
            "cn=bad-two-entries,dc=example,dc=com",
        )

    def test_lookup_fail_emptyError(self):
        writeFile(os.path.join(self.example.path, "cn=bad-empty.ldif"), b"")
        self.assertRaises(
            ldiftree.LDIFTreeEntryContainsNoEntries,
            self.example.lookup,
            "cn=bad-empty,dc=example,dc=com",
        )

    def test_lookup_deep(self):
        dn = "cn=bar,ou=metasyntactic,dc=example,dc=com"
        d = self.root.lookup(dn)
        d.addCallback(self._cb_test_lookup_deep)
        return d

    def _cb_test_lookup_deep(self, r):
        self.assertEqual(r, self.bar)

    def test_delete_root(self):
        d = self.root.delete()

        def eb(fail):
            fail.trap(ldiftree.LDAPCannotRemoveRootError)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_delete_nonLeaf(self):
        d = self.meta.delete()

        def eb(fail):
            fail.trap(ldaperrors.LDAPNotAllowedOnNonLeaf)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_delete(self):
        d = self.foo.delete()
        d.addCallback(self._cb_test_delete_1)
        return d

    def _cb_test_delete_1(self, r):
        self.assertEqual(r, self.foo)
        d = self.meta.children()
        d.addCallback(self._cb_test_delete_2)
        return d

    def _cb_test_delete_2(self, r):
        self.assertEqual(r, [self.bar])

    def test_deleteChild(self):
        d = self.meta.deleteChild("cn=bar")
        d.addCallback(self._cb_test_deleteChild_1)
        return d

    def _cb_test_deleteChild_1(self, r):
        self.assertEqual(r, self.bar)
        d = self.meta.children()
        d.addCallback(self._cb_test_deleteChild_2)
        return d

    def _cb_test_deleteChild_2(self, r):
        self.assertEqual(r, [self.foo])

    def test_deleteChild_NonExisting(self):
        d = self.root.deleteChild("cn=not-exist")

        def eb(fail):
            fail.trap(ldaperrors.LDAPNoSuchObject)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_setPassword(self):
        self.foo.setPassword(b"s3krit", salt=b"\xf2\x4a")
        self.failUnless("userPassword" in self.foo)
        self.assertEqual(
            self.foo["userPassword"], [b"{SSHA}0n/Iw1NhUOKyaI9gm9v5YsO3ZInySg=="]
        )

    def test_setPassword_noSalt(self):
        self.foo.setPassword(b"s3krit")
        self.failUnless("userPassword" in self.foo)
        d = self.foo.bind("s3krit")
        d.addCallback(self.assertIdentical, self.foo)
        d.addCallback(lambda _: self.foo.bind("s4krit"))

        def eb(fail):
            fail.trap(ldaperrors.LDAPInvalidCredentials)

        d.addCallbacks(testutil.mustRaise, eb)
        return d

    def test_diffTree_self(self):
        d = self.root.diffTree(self.root)
        d.addCallback(self.assertEqual, [])
        return d

    def test_diffTree_copy(self):
        otherDir = self.mktemp()
        shutil.copytree(self.tree, otherDir)
        other = ldiftree.LDIFTreeEntry(otherDir)
        d = self.root.diffTree(other)
        d.addCallback(self.assertEqual, [])
        return d

    def test_diffTree_addChild(self):
        otherDir = self.mktemp()
        shutil.copytree(self.tree, otherDir)
        other = ldiftree.LDIFTreeEntry(otherDir)
        e = entry.BaseLDAPEntry(dn="cn=foo,dc=example,dc=com")
        d = ldiftree.put(otherDir, e)

        def cb1(dummy):
            return other.lookup("cn=foo,dc=example,dc=com")

        d.addCallback(cb1)

        def cb2(r):
            d = self.root.diffTree(other)
            d.addCallback(self.assertEqual, [delta.AddOp(r)])
            return d

        d.addCallback(cb2)
        return d

    def test_diffTree_delChild(self):
        otherDir = self.mktemp()
        shutil.copytree(self.tree, otherDir)
        other = ldiftree.LDIFTreeEntry(otherDir)

        d = other.lookup("ou=empty,dc=example,dc=com")

        def cb1(otherEmpty):
            return otherEmpty.delete()

        d.addCallback(cb1)

        def cb2(dummy):
            return self.root.diffTree(other)

        d.addCallback(cb2)

        def cb3(got):
            self.assertEqual(got, [delta.DeleteOp(self.empty)])

        d.addCallback(cb3)
        return d

    @util.fromCoroutineFunction
    async def test_diffTree_edit_failure(self):
        otherDir = self.mktemp()
        shutil.copytree(self.tree, otherDir)
        other = ldiftree.LDIFTreeEntry(otherDir)

        otherEmpty = await other.lookup("ou=empty,dc=example,dc=com")
        otherEmpty["foo"] = ["bar"]
        shutil.rmtree(otherDir)
        observer = testing.EventLoggingObserver.createWithCleanup(self, log)
        self.assertFalse(await otherEmpty.commit())
        self.assertEqual(
            observer[0]["log_text"],
            "[ERROR] Could not commit entry: ou=empty,dc=example,dc=com.",
        )

    @skipIfWindows
    def test_diffTree_edit(self):
        otherDir = self.mktemp()
        shutil.copytree(self.tree, otherDir)
        other = ldiftree.LDIFTreeEntry(otherDir)

        d = other.lookup("ou=empty,dc=example,dc=com")

        def cb1(otherEmpty):
            otherEmpty["foo"] = ["bar"]
            return otherEmpty.commit()

        d.addCallback(cb1)

        def cb2(dummy):
            return self.root.diffTree(other)

        d.addCallback(cb2)

        def cb3(got):
            self.assertEqual(
                got,
                [
                    delta.ModifyOp(
                        self.empty.dn,
                        [delta.Add(b"foo", [b"bar"])],
                    ),
                ],
            )

        d.addCallback(cb3)
        return d

    @skipIfWindows
    def test_move_noChildren_sameSuperior(self):
        d = self.empty.move("ou=moved,dc=example,dc=com")

        def getChildren(dummy):
            return self.example.children()

        d.addCallback(getChildren)
        d.addCallback(set)
        d.addCallback(
            self.assertEqual,
            {
                self.meta,
                BaseLDAPEntry(
                    dn="ou=moved,dc=example,dc=com",
                    attributes={
                        b"objectClass": [b"a", b"b"],
                        b"ou": [b"moved"],
                    },
                ),
                self.oneChild,
            },
        )
        return d

    @skipIfWindows
    def test_move_children_sameSuperior(self):
        d = self.meta.move("ou=moved,dc=example,dc=com")

        def getChildren(dummy):
            return self.example.children()

        d.addCallback(getChildren)
        d.addCallback(set)
        d.addCallback(
            self.assertEqual,
            {
                BaseLDAPEntry(
                    dn="ou=moved,dc=example,dc=com",
                    attributes={
                        b"objectClass": [b"a", b"b"],
                        b"ou": [b"moved"],
                    },
                ),
                self.empty,
                self.oneChild,
            },
        )
        return d

    @skipIfWindows
    def test_move_noChildren_newSuperior(self):
        d = self.empty.move("ou=moved,ou=oneChild,dc=example,dc=com")

        def getChildren(dummy):
            return self.example.children()

        d.addCallback(getChildren)
        d.addCallback(set)
        d.addCallback(
            self.assertEqual,
            {
                self.meta,
                self.oneChild,
            },
        )

        def getChildren2(dummy):
            return self.oneChild.children()

        d.addCallback(getChildren2)
        d.addCallback(set)
        d.addCallback(
            self.assertEqual,
            {
                self.theChild,
                BaseLDAPEntry(
                    dn="ou=moved,ou=oneChild,dc=example,dc=com",
                    attributes={
                        b"objectClass": [b"a", b"b"],
                        b"ou": [b"moved"],
                    },
                ),
            },
        )
        return d

    @skipIfWindows
    def test_move_children_newSuperior(self):
        d = self.meta.move("ou=moved,ou=oneChild,dc=example,dc=com")

        def getChildren(dummy):
            return self.example.children()

        d.addCallback(getChildren)
        d.addCallback(set)
        d.addCallback(
            self.assertEqual,
            {
                self.empty,
                self.oneChild,
            },
        )

        def getChildren2(dummy):
            return self.oneChild.children()

        d.addCallback(getChildren2)
        d.addCallback(set)
        d.addCallback(
            self.assertEqual,
            {
                self.theChild,
                BaseLDAPEntry(
                    dn="ou=moved,ou=oneChild,dc=example,dc=com",
                    attributes={
                        b"objectClass": [b"a", b"b"],
                        b"ou": [b"moved"],
                    },
                ),
            },
        )
        return d

    def testCompareOtherTypes(self):
        """
        It can't be compared with other types.
        """
        with self.assertRaises(TypeError):
            self.example < object()

        with self.assertRaises(TypeError):
            self.example > object()

    def testCompareGreater(self):
        """
        It is compared with other entries based on DN, where child is
        greater than the parent.
        """
        self.assertTrue(self.oneChild > self.example)
        self.assertFalse(self.example > self.oneChild)

    def testCompareLess(self):
        """
        It is compared with other entries based on DN, where parent is
        less than the child.
        """
        self.assertTrue(self.example < self.oneChild)
        self.assertFalse(self.oneChild < self.example)



"""
Test cases for ldaptor.protocols.ldap.proxybase module.
"""
from functools import partial
import itertools

from twisted.internet import error, defer
from twisted.internet.task import Clock
from twisted.trial import unittest
from twisted.test import proto_helpers

from ldaptor.protocols.ldap import proxybase, ldaperrors
from ldaptor.protocols import pureldap
from ldaptor import testutil


class RequestInterceptingProxy(proxybase.ProxyBase):
    """
    A test LDAP proxy that does not forward requests but instead
    responses with pre-determined responses.
    """

    responses = [pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode)]

    def handleBeforeForwardRequest(self, request, controls, reply):
        """
        Don't forward the message to the proxied service-- instead
        reply with predetermined responses.
        """
        for response in self.responses:
            reply(response)
        return defer.succeed(None)


class ResponseInterceptingProxy(proxybase.ProxyBase):
    """
    A test LDAP proxy that intercepts and modifies search results.
    """

    delays = itertools.cycle([1, 5])
    new_attrib = ("frotz", "xyzzy")

    def handleProxiedResponse(self, response, request, controls):
        """
        If the response is an LDAPSearchResultEntry, modify
        the attribute list
        """
        if isinstance(response, pureldap.LDAPSearchResultEntry):
            key, value = self.new_attrib
            response.attributes.append((key, [value]))
        d = defer.Deferred()
        d.addCallback(self._afterDelay)
        self.reactor.callLater(next(self.delays), d.callback, response)
        return d

    def _afterDelay(self, response):
        return response


class WontConnectError(Exception):
    pass


def failToConnectToServer(reactor, delay=0):
    d = defer.Deferred()

    def onConnect():
        d.errback(fail=WontConnectError("Test LDAP client refuses to connect."))

    reactor.callLater(delay, onConnect)
    return d


class ProxyBase(unittest.TestCase):
    def createServer(self, *responses, **kwds):
        """
        Create a server for each test.
        """
        protocol = kwds.get("protocol", proxybase.ProxyBase)
        clock = Clock()
        clock = kwds.get("clock", clock)
        server = protocol()
        clientTestDriver = testutil.LDAPClientTestDriver(*responses)

        def simulateConnectToServer():
            d = defer.Deferred()

            def onConnect():
                clientTestDriver.connectionMade()
                d.callback(clientTestDriver)

            clock.callLater(0, onConnect)
            return d

        clientConnector = kwds.get("clientConnector", simulateConnectToServer)
        server.clientConnector = clientConnector
        server.clientTestDriver = clientTestDriver
        server.transport = proto_helpers.StringTransport()
        server.reactor = clock
        server.connectionMade()
        return server

    def test_bind(self):
        """
        When binding to the server an `LDAPBindResponse` with a successful
        result code.is written to the transport.
        """
        server = self.createServer([pureldap.LDAPBindResponse(resultCode=0)])
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=4).toWire()
        )
        server.reactor.advance(1)
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(resultCode=0), id=4
            ).toWire(),
        )

    def test_search(self):
        """
        When performing an LDAP search against the server; the search results and
        a single "search done" response is written to the transport.
        """
        server = self.createServer(
            [pureldap.LDAPBindResponse(resultCode=0)],
            [
                pureldap.LDAPSearchResultEntry(
                    "cn=foo,dc=example,dc=com", [("a", ["b"])]
                ),
                pureldap.LDAPSearchResultEntry(
                    "cn=bar,dc=example,dc=com", [("b", ["c"])]
                ),
                pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode),
            ],
        )
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=2).toWire()
        )
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPSearchRequest(), id=3).toWire()
        )
        server.reactor.advance(1)
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(pureldap.LDAPBindResponse(resultCode=0), id=2).toWire()
            + pureldap.LDAPMessage(
                pureldap.LDAPSearchResultEntry(
                    "cn=foo,dc=example,dc=com", [("a", ["b"])]
                ),
                id=3,
            ).toWire()
            + pureldap.LDAPMessage(
                pureldap.LDAPSearchResultEntry(
                    "cn=bar,dc=example,dc=com", [("b", ["c"])]
                ),
                id=3,
            ).toWire()
            + pureldap.LDAPMessage(
                pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode), id=3
            ).toWire(),
        )

    def test_unbind_clientUnbinds(self):
        """
        The server disconnects from the client gracefully when the
        client signals its intent to unbind.
        """
        server = self.createServer(
            [pureldap.LDAPBindResponse(resultCode=0)],
            [],
        )
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=2).toWire()
        )
        server.reactor.advance(1)
        client = server.client
        client.assertSent(pureldap.LDAPBindRequest())
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(resultCode=0), id=2
            ).toWire(),
        )
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPUnbindRequest(), id=3).toWire()
        )
        server.connectionLost(error.ConnectionDone)
        server.reactor.advance(1)
        client.assertSent(pureldap.LDAPBindRequest(), pureldap.LDAPUnbindRequest())
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(resultCode=0), id=2
            ).toWire(),
        )

    def test_unbind_clientEOF(self):
        """
        The server disconects correctly when the client terminates the
        connection without sending an unbind request.
        """
        server = self.createServer([pureldap.LDAPBindResponse(resultCode=0)])
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=2).toWire()
        )
        server.reactor.advance(1)
        client = server.client
        client.assertSent(pureldap.LDAPBindRequest())
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(resultCode=0), id=2
            ).toWire(),
        )
        server.connectionLost(error.ConnectionDone)
        server.reactor.advance(1)
        client.assertSent(
            pureldap.LDAPBindRequest(), "fake-unbind-by-LDAPClientTestDriver"
        )
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(resultCode=0), id=2
            ).toWire(),
        )

    def test_intercepted_search_request(self):
        """
        When performing an LDAP search against the server; the requests are
        intercepted and custom responses are written to the transport.
        """
        server = self.createServer(
            [pureldap.LDAPBindResponse(resultCode=0)],
            [
                pureldap.LDAPSearchResultEntry(
                    "cn=foo,dc=example,dc=com", [("a", ["b"])]
                ),
                pureldap.LDAPSearchResultEntry(
                    "cn=bar,dc=example,dc=com", [("b", ["c"])]
                ),
                pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode),
            ],
            protocol=RequestInterceptingProxy,
        )
        server.responses = [
            pureldap.LDAPSearchResultEntry(
                "cn=xyzzy,dc=example,dc=com", [("frobnitz", ["zork"])]
            ),
            pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode),
        ]
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPSearchRequest(), id=1).toWire()
        )
        server.reactor.advance(1)
        self.assertEqual(len(server.clientTestDriver.sent), 0)
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPSearchResultEntry(
                    "cn=xyzzy,dc=example,dc=com", [("frobnitz", ["zork"])]
                ),
                id=1,
            ).toWire()
            + pureldap.LDAPMessage(
                pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode), id=1
            ).toWire(),
        )

    def test_intercepted_search_response(self):
        """
        When performing an LDAP search against the server; the search results are
        intercepted and modified by the proxy.
        """
        server = self.createServer(
            [pureldap.LDAPBindResponse(resultCode=0)],
            [
                pureldap.LDAPSearchResultEntry(
                    "cn=foo,dc=example,dc=com", [("a", ["b"])]
                ),
                pureldap.LDAPSearchResultEntry(
                    "cn=bar,dc=example,dc=com", [("b", ["c"])]
                ),
                pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode),
            ],
            protocol=ResponseInterceptingProxy,
        )
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=2).toWire()
        )
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPSearchRequest(), id=3).toWire()
        )
        server.reactor.advance(1)
        server.reactor.advance(5)
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(pureldap.LDAPBindResponse(resultCode=0), id=2).toWire()
            + pureldap.LDAPMessage(
                pureldap.LDAPSearchResultEntry(
                    "cn=foo,dc=example,dc=com", [("a", ["b"]), ("frotz", ["xyzzy"])]
                ),
                id=3,
            ).toWire()
            + pureldap.LDAPMessage(
                pureldap.LDAPSearchResultEntry(
                    "cn=bar,dc=example,dc=com", [("b", ["c"]), ("frotz", ["xyzzy"])]
                ),
                id=3,
            ).toWire()
            + pureldap.LDAPMessage(
                pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode), id=3
            ).toWire(),
        )

    def test_cannot_connect_to_proxied_server_no_pending_requests(self):
        """
        When making a request and the proxy cannot connect to the proxied server, the
        connection is terminated.
        """
        clock = Clock()
        connector = partial(failToConnectToServer, clock)
        server = self.createServer([], clientConnector=connector, clock=clock)
        self.assertEqual(connector, server.clientConnector)
        server.reactor.advance(1)
        self.assertEqual(server.transport.value(), b"")

    def test_cannot_connect_to_proxied_server_pending_requests(self):
        """
        When making a request and the proxy cannot connect to the proxied server,
        pending BIND and startTLS requests are replied to and the connection
        is closed.
        """
        clock = Clock()
        connector = partial(failToConnectToServer, clock)
        server = self.createServer([], clientConnector=connector, clock=clock)
        self.assertEqual(connector, server.clientConnector)
        server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=4).toWire()
        )
        server.reactor.advance(2)
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(resultCode=52), id=4
            ).toWire(),
        )

    def test_health_check_closes_connection_to_proxied_server(self):
        """
        When the client disconnects immediately and before the connection to the proxied server has
        been established, the proxy terminates the connection to the proxied server.
        Messages sent by the client are discarded.
        """
        request = pureldap.LDAPBindRequest()
        message = pureldap.LDAPMessage(request, id=4)
        server = self.createServer()
        # Send a message, message is queued
        server.dataReceived(message.toWire())
        self.assertEqual(len(server.queuedRequests), 1)
        self.assertEqual(server.queuedRequests[0][0], request)
        # Lose connection, message is discarded
        server.connectionLost(error.ConnectionDone)
        server.reactor.advance(1)
        self.assertIsNone(server.client)
        self.assertFalse(server.clientTestDriver.connected)
        self.assertEqual(server.queuedRequests, [])



"""
Test cases for ldaptor.protocols.ldap.ldapserver module.
"""
import base64
import types

from twisted.internet import address, protocol, testing
from twisted.python import components, log
from twisted.test import proto_helpers
from twisted.trial import unittest

from ldaptor import inmemory, interfaces, schema, delta, entry
from ldaptor.protocols.ldap import ldapserver, ldapclient, ldaperrors, fetchschema
from ldaptor.protocols import pureldap, pureber
from ldaptor.test import util, test_schema
from ldaptor._encoder import to_bytes


def observeCommits(entry):
    commits = []
    bound_commit = entry.commit

    def observe(v):
        commits.append(v)
        return v

    def commit(self):
        d = bound_commit()
        d.addBoth(observe)
        return d

    entry.commit = types.MethodType(commit, entry)
    return commits


class LDAPServerTest(unittest.TestCase):
    def setUp(self):
        self.root = inmemory.ReadOnlyInMemoryLDAPEntry(
            dn="dc=example,dc=com", attributes={"dc": "example"}
        )
        self.stuff = self.root.addChild(
            rdn="ou=stuff",
            attributes={
                b"objectClass": [b"a", b"b"],
                b"ou": [b"stuff"],
            },
        )
        self.thingie = self.stuff.addChild(
            rdn="cn=thingie",
            attributes={
                "objectClass": ["a", "b"],
                "cn": ["thingie"],
            },
        )
        self.another = self.stuff.addChild(
            rdn="cn=another",
            attributes={
                "objectClass": ["a", "b"],
                "cn": ["another"],
            },
        )

        # Add Users Subtree
        self.users = self.root.addChild(
            rdn="ou=People",
            attributes={"objectClass": ["top", "organizationalunit"], "ou": ["People"]},
        )

        self.users.addChild(
            rdn="uid=kthompson",
            attributes={"objectClass": ["top", "inetOrgPerson"], "uid": ["kthompson"]},
        )

        self.users.addChild(
            rdn="uid=bgates",
            attributes={"objectClass": ["top", "inetOrgPerson"], "uid": ["bgates"]},
        )

        # Add Groups Subtree
        self.groups = self.root.addChild(
            rdn="ou=Groups",
            attributes={"objectClass": ["top", "organizationalunit"], "ou": ["Groups"]},
        )

        self.groups.addChild(
            rdn="cn=unix",
            attributes={
                "uniquemember": ["uid=kthompson,ou=People,dc=example,dc=com"],
                "objectClass": ["top", "groupOfUniqueNames"],
                "cn": ["unix"],
            },
        )

        server = ldapserver.LDAPServer()
        server.factory = self.root
        server.transport = proto_helpers.StringTransport()
        server.connectionMade()
        self.server = server

    def _makeResultList(self, s):
        berdecoder = pureldap.LDAPBERDecoderContext_TopLevel(
            inherit=pureldap.LDAPBERDecoderContext_LDAPMessage(
                fallback=pureldap.LDAPBERDecoderContext(
                    fallback=pureber.BERDecoderContext()
                ),
                inherit=pureldap.LDAPBERDecoderContext(
                    fallback=pureber.BERDecoderContext()
                ),
            )
        )
        buffer = s
        value = []
        while 1:
            o, bytes = pureber.berDecodeObject(berdecoder, buffer)
            buffer = buffer[bytes:]
            if not o:
                break
            value.append(o.toWire())
        return value

    def makeSearch(
        self,
        baseObject=None,
        scope=None,
        derefAliases=None,
        sizeLimit=None,
        timeLimit=None,
        typesOnly=None,
        filter=None,
        attributes=None,
        tag=None,
    ):
        """Shortcut for sending LDAPSearchRequest to the test server"""
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPSearchRequest(
                    baseObject=baseObject,
                    scope=scope,
                    derefAliases=derefAliases,
                    sizeLimit=sizeLimit,
                    timeLimit=timeLimit,
                    typesOnly=typesOnly,
                    filter=filter,
                    attributes=attributes,
                    tag=tag,
                ),
                id=2,
            ).toWire()
        )

    def assertSearchResults(self, results=None, resultCode=0):
        """
        Shortcut for checking results returned by test server on LDAPSearchRequest.
        Results must be prepared as a list of dictionaries with 'objectName' and 'attributes' keys
        """
        if results is None:
            results = []

        messages = []

        for result in results:
            message = pureldap.LDAPMessage(
                pureldap.LDAPSearchResultEntry(
                    objectName=result["objectName"], attributes=result["attributes"]
                ),
                id=2,
            )
            messages.append(message)

        messages.append(
            pureldap.LDAPMessage(
                pureldap.LDAPSearchResultDone(resultCode=resultCode), id=2
            )
        )
        self.assertCountEqual(
            self._makeResultList(self.server.transport.value()),
            [msg.toWire() for msg in messages],
        )

    def test_bind(self):
        self.server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=4).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(resultCode=0), id=4
            ).toWire(),
        )

    def test_bind_success(self):
        self.thingie["userPassword"] = [
            "{SSHA}yVLLj62rFf3kDAbzwEU0zYAVvbWrze8="
        ]  # "secret"
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(
                    dn="cn=thingie,ou=stuff,dc=example,dc=com", auth=b"secret"
                ),
                id=4,
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=0, matchedDN="cn=thingie,ou=stuff,dc=example,dc=com"
                ),
                id=4,
            ).toWire(),
        )

    def test_bind_invalidCredentials_badPassword(self):
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(
                    dn="cn=thingie,ou=stuff,dc=example,dc=com", auth=b"invalid"
                ),
                id=734,
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.LDAPInvalidCredentials.resultCode
                ),
                id=734,
            ).toWire(),
        )

    def test_bind_invalidCredentials_nonExisting(self):
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(
                    dn="cn=non-existing,dc=example,dc=com", auth=b"invalid"
                ),
                id=78,
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.LDAPInvalidCredentials.resultCode
                ),
                id=78,
            ).toWire(),
        )

    def test_bind_badVersion_1_anonymous(self):
        self.server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPBindRequest(version=1), id=32).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.LDAPProtocolError.resultCode,
                    errorMessage="Version 1 not supported",
                ),
                id=32,
            ).toWire(),
        )

    def test_bind_badVersion_2_anonymous(self):
        self.server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPBindRequest(version=2), id=32).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.LDAPProtocolError.resultCode,
                    errorMessage="Version 2 not supported",
                ),
                id=32,
            ).toWire(),
        )

    def test_bind_badVersion_4_anonymous(self):
        self.server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPBindRequest(version=4), id=32).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.LDAPProtocolError.resultCode,
                    errorMessage="Version 4 not supported",
                ),
                id=32,
            ).toWire(),
        )

    def test_bind_badVersion_4_nonExisting(self):
        # TODO make a test just like this one that would pass authentication
        # if version was correct, to ensure we don't leak that info either.
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(
                    version=4, dn="cn=non-existing,dc=example,dc=com", auth=b"invalid"
                ),
                id=11,
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.LDAPProtocolError.resultCode,
                    errorMessage="Version 4 not supported",
                ),
                id=11,
            ).toWire(),
        )

    def test_unbind(self):
        self.server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPUnbindRequest(), id=7).toWire()
        )
        self.assertEqual(self.server.transport.value(), b"")

    def test_compare_outOfTree(self):
        dn = "dc=invalid"
        attribute_desc = pureldap.LDAPString("objectClass")
        attribute_value = pureldap.LDAPString("groupOfUniqueNames")
        ava = pureldap.LDAPAttributeValueAssertion(attribute_desc, attribute_value)

        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPCompareRequest(entry=dn, ava=ava), id=2
            ).toWire()
        )

        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPCompareResponse(
                    resultCode=ldaperrors.LDAPNoSuchObject.resultCode
                ),
                id=2,
            ).toWire(),
        )

    def test_compare_inGroup(self):
        dn = "cn=unix,ou=Groups,dc=example,dc=com"
        attribute_desc = pureldap.LDAPString("uniquemember")
        attribute_value = pureldap.LDAPString(
            "uid=kthompson,ou=People,dc=example,dc=com"
        )
        ava = pureldap.LDAPAttributeValueAssertion(attribute_desc, attribute_value)

        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPCompareRequest(entry=dn, ava=ava), id=2
            ).toWire()
        )

        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPCompareResponse(
                    resultCode=ldaperrors.LDAPCompareTrue.resultCode
                ),
                id=2,
            ).toWire(),
        )

    def test_compare_notInGroup(self):
        dn = "cn=unix,ou=Groups,dc=example,dc=com"
        attribute_desc = pureldap.LDAPString("uniquemember")
        attribute_value = pureldap.LDAPString("uid=bgates,ou=People,dc=example,dc=com")
        ava = pureldap.LDAPAttributeValueAssertion(attribute_desc, attribute_value)

        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPCompareRequest(entry=dn, ava=ava), id=2
            ).toWire()
        )

        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPCompareResponse(
                    resultCode=ldaperrors.LDAPCompareFalse.resultCode
                ),
                id=2,
            ).toWire(),
        )

    def test_search_outOfTree(self):
        """Attempt to get nonexistent DN results in noSuchObject error response"""
        self.makeSearch(baseObject="dc=invalid")
        self.assertSearchResults(resultCode=ldaperrors.LDAPNoSuchObject.resultCode)

    def test_search_matchAll_oneResult(self):
        """Searching for a single object with receiving all its attributes"""
        self.makeSearch(baseObject="cn=thingie,ou=stuff,dc=example,dc=com")
        self.assertSearchResults(
            [
                {
                    "objectName": "cn=thingie,ou=stuff,dc=example,dc=com",
                    "attributes": [
                        ("objectClass", ["a", "b"]),
                        ("cn", ["thingie"]),
                    ],
                }
            ]
        )

    def test_search_matchAll_oneResult_filtered(self):
        """Searching for a single object with receiving a specified set of its attributes"""
        self.makeSearch(
            baseObject="cn=thingie,ou=stuff,dc=example,dc=com", attributes=["cn"]
        )
        self.assertSearchResults(
            [
                {
                    "objectName": "cn=thingie,ou=stuff,dc=example,dc=com",
                    "attributes": [
                        ("cn", ["thingie"]),
                    ],
                }
            ]
        )

    def test_search_matchAll_oneResult_filteredNoAttribsRemaining(self):
        """
        Attempt to search an existing object with a set of nonexistent attributes
        results in a successful response with no attributes
        """
        self.makeSearch(
            baseObject="cn=thingie,ou=stuff,dc=example,dc=com", attributes=["xyzzy"]
        )
        self.assertSearchResults(
            [
                {
                    "objectName": "cn=thingie,ou=stuff,dc=example,dc=com",
                    "attributes": [],
                },
            ],
        )

    def test_search_matchAll_manyResults(self):
        """Searching for a tree object with receiving it and all its children (default scope)"""
        self.makeSearch(baseObject="ou=stuff,dc=example,dc=com")
        self.assertSearchResults(
            [
                {
                    "objectName": "ou=stuff,dc=example,dc=com",
                    "attributes": [
                        ("objectClass", ["a", "b"]),
                        ("ou", ["stuff"]),
                    ],
                },
                {
                    "objectName": "cn=another,ou=stuff,dc=example,dc=com",
                    "attributes": [
                        ("objectClass", ["a", "b"]),
                        ("cn", ["another"]),
                    ],
                },
                {
                    "objectName": "cn=thingie,ou=stuff,dc=example,dc=com",
                    "attributes": [
                        ("objectClass", ["a", "b"]),
                        ("cn", ["thingie"]),
                    ],
                },
            ]
        )

    def test_search_scope_oneLevel(self):
        """Searching for a tree object with receiving its children but without parent itself"""
        self.makeSearch(
            baseObject="ou=stuff,dc=example,dc=com",
            scope=pureldap.LDAP_SCOPE_singleLevel,
        )
        self.assertSearchResults(
            [
                {
                    "objectName": "cn=thingie,ou=stuff,dc=example,dc=com",
                    "attributes": [
                        ("objectClass", ["a", "b"]),
                        ("cn", ["thingie"]),
                    ],
                },
                {
                    "objectName": "cn=another,ou=stuff,dc=example,dc=com",
                    "attributes": [
                        ("objectClass", ["a", "b"]),
                        ("cn", ["another"]),
                    ],
                },
            ]
        )

    def test_search_scope_wholeSubtree(self):
        """
        Searching for a tree object with receiving it and all its children.
        This is a default behavior but here it is explicitly specified.
        """
        self.makeSearch(
            baseObject="ou=stuff,dc=example,dc=com",
            scope=pureldap.LDAP_SCOPE_wholeSubtree,
        )
        self.assertSearchResults(
            [
                {
                    "objectName": "ou=stuff,dc=example,dc=com",
                    "attributes": [
                        ("objectClass", ["a", "b"]),
                        ("ou", ["stuff"]),
                    ],
                },
                {
                    "objectName": "cn=another,ou=stuff,dc=example,dc=com",
                    "attributes": [
                        ("objectClass", ["a", "b"]),
                        ("cn", ["another"]),
                    ],
                },
                {
                    "objectName": "cn=thingie,ou=stuff,dc=example,dc=com",
                    "attributes": [
                        ("objectClass", ["a", "b"]),
                        ("cn", ["thingie"]),
                    ],
                },
            ]
        )

    def test_search_scope_baseObject(self):
        """Searching for a tree object with receiving it without its children"""
        self.makeSearch(
            baseObject="ou=stuff,dc=example,dc=com",
            scope=pureldap.LDAP_SCOPE_baseObject,
        )
        self.assertSearchResults(
            [
                {
                    "objectName": "ou=stuff,dc=example,dc=com",
                    "attributes": [
                        ("objectClass", ["a", "b"]),
                        ("ou", ["stuff"]),
                    ],
                }
            ]
        )

    def test_search_all_attributes(self):
        """
        Search request with the list of attributes passed as '*'
        returns objects with all their attributes
        """
        self.makeSearch(
            baseObject="cn=thingie,ou=stuff,dc=example,dc=com", attributes=["*"]
        )
        self.assertSearchResults(
            [
                {
                    "objectName": "cn=thingie,ou=stuff,dc=example,dc=com",
                    "attributes": [
                        ("objectClass", ["a", "b"]),
                        ("cn", ["thingie"]),
                    ],
                }
            ]
        )

    def test_rootDSE(self):
        """Searching for a root object"""
        self.makeSearch(
            baseObject="",
            scope=pureldap.LDAP_SCOPE_baseObject,
            filter=pureldap.LDAPFilter_present("objectClass"),
        )
        self.assertSearchResults(
            [
                {
                    "objectName": "",
                    "attributes": [
                        ("supportedLDAPVersion", ["3"]),
                        ("namingContexts", ["dc=example,dc=com"]),
                        (
                            "supportedExtension",
                            [pureldap.LDAPPasswordModifyRequest.oid],
                        ),
                    ],
                }
            ]
        )

    def test_delete(self):
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPDelRequest(self.thingie.dn.getText()), id=2
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(pureldap.LDAPDelResponse(resultCode=0), id=2).toWire(),
        )
        d = self.stuff.children()
        d.addCallback(lambda actual: self.assertCountEqual(actual, [self.another]))
        return d

    def test_add_success(self):
        dn = "cn=new,ou=stuff,dc=example,dc=com"
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPAddRequest(
                    entry=dn,
                    attributes=[
                        (
                            pureldap.LDAPAttributeDescription("objectClass"),
                            pureber.BERSet(
                                value=[pureldap.LDAPAttributeValue("something")]
                            ),
                        )
                    ],
                ),
                id=2,
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPAddResponse(resultCode=ldaperrors.Success.resultCode), id=2
            ).toWire(),
        )
        # tree changed
        d = self.stuff.children()
        d.addCallback(
            lambda actual: self.assertCountEqual(
                actual,
                [
                    self.thingie,
                    self.another,
                    inmemory.ReadOnlyInMemoryLDAPEntry(
                        b"cn=new,ou=stuff,dc=example,dc=com",
                        {b"objectClass": [b"something"]},
                    ),
                ],
            )
        )
        return d

    def test_add_fail_existsAlready(self):
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPAddRequest(
                    entry=self.thingie.dn.getText(),
                    attributes=[
                        (
                            pureldap.LDAPAttributeDescription("objectClass"),
                            pureber.BERSet(
                                value=[
                                    pureldap.LDAPAttributeValue("something"),
                                ]
                            ),
                        )
                    ],
                ),
                id=2,
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPAddResponse(
                    resultCode=ldaperrors.LDAPEntryAlreadyExists.resultCode,
                    errorMessage=self.thingie.dn.getText(),
                ),
                id=2,
            ).toWire(),
        )
        # tree did not change
        d = self.stuff.children()
        d.addCallback(
            lambda actual: self.assertCountEqual(actual, [self.thingie, self.another])
        )
        return d

    def test_modifyDN_rdnOnly_deleteOldRDN_success(self):
        newrdn = "cn=thingamagic"
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPModifyDNRequest(
                    entry=self.thingie.dn.getText(), newrdn=newrdn, deleteoldrdn=True
                ),
                id=2,
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPModifyDNResponse(resultCode=ldaperrors.Success.resultCode),
                id=2,
            ).toWire(),
        )
        # tree changed
        d = self.stuff.children()
        d.addCallback(
            lambda actual: self.assertCountEqual(
                actual,
                [
                    inmemory.ReadOnlyInMemoryLDAPEntry(
                        "%s,ou=stuff,dc=example,dc=com" % newrdn,
                        {"objectClass": ["a", "b"], "cn": ["thingamagic"]},
                    ),
                    self.another,
                ],
            )
        )
        return d

    def test_modify(self):
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPModifyRequest(
                    self.stuff.dn.getText(),
                    modification=[
                        delta.Add("foo", ["bar"]).asLDAP(),
                    ],
                ),
                id=2,
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPModifyResponse(resultCode=ldaperrors.Success.resultCode),
                id=2,
            ).toWire(),
        )
        # tree changed
        self.assertEqual(
            self.stuff,
            inmemory.ReadOnlyInMemoryLDAPEntry(
                "ou=stuff,dc=example,dc=com",
                {b"objectClass": [b"a", b"b"], b"ou": [b"stuff"], b"foo": [b"bar"]},
            ),
        )

    def test_extendedRequest_unknown(self):
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPExtendedRequest(
                    requestName="42.42.42", requestValue="foo"
                ),
                id=2,
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPExtendedResponse(
                    resultCode=ldaperrors.LDAPProtocolError.resultCode,
                    errorMessage="Unknown extended request: 42.42.42",
                ),
                id=2,
            ).toWire(),
        )

    def test_passwordModify_notBound(self):
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPPasswordModifyRequest(
                    userIdentity="cn=thingie,ou=stuff,dc=example,dc=com",
                    newPasswd="hushhush",
                ),
                id=2,
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPExtendedResponse(
                    resultCode=ldaperrors.LDAPStrongAuthRequired.resultCode,
                    responseName=pureldap.LDAPPasswordModifyRequest.oid,
                ),
                id=2,
            ).toWire(),
        )

    def test_passwordModify_simple(self):
        commits = observeCommits(self.thingie)
        # first bind to some entry
        self.thingie["userPassword"] = [
            "{SSHA}yVLLj62rFf3kDAbzwEU0zYAVvbWrze8="
        ]  # "secret"
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(
                    dn="cn=thingie,ou=stuff,dc=example,dc=com", auth=b"secret"
                ),
                id=4,
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=0, matchedDN="cn=thingie,ou=stuff,dc=example,dc=com"
                ),
                id=4,
            ).toWire(),
        )
        self.server.transport.clear()
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPPasswordModifyRequest(
                    userIdentity="cn=thingie,ou=stuff,dc=example,dc=com",
                    newPasswd="hushhush",
                ),
                id=2,
            ).toWire()
        )
        self.assertListEqual(commits, [True], "Server never committed data.")
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPExtendedResponse(
                    resultCode=ldaperrors.Success.resultCode,
                    responseName=pureldap.LDAPPasswordModifyRequest.oid,
                ),
                id=2,
            ).toWire(),
        )
        # tree changed
        secrets = self.thingie.get("userPassword", [])
        self.assertEqual(len(secrets), 1)
        for secret in secrets:
            self.assertEqual(secret[: len(b"{SSHA}")], b"{SSHA}")
            raw = base64.decodestring(secret[len(b"{SSHA}") :])
            salt = raw[20:]
            self.assertEqual(entry.sshaDigest(b"hushhush", salt), secret)

    def test_passwordModify_someoneElse(self):
        commits = observeCommits(self.thingie)
        # first bind to some entry
        userPassword = b"{SSHA}yVLLj62rFf3kDAbzwEU0zYAVvbWrze8="  # secret
        self.thingie["userPassword"] = [userPassword]
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(
                    dn="cn=thingie,ou=stuff,dc=example,dc=com", auth=b"secret"
                ),
                id=4,
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=0, matchedDN="cn=thingie,ou=stuff,dc=example,dc=com"
                ),
                id=4,
            ).toWire(),
        )
        self.server.transport.clear()
        observer = testing.EventLoggingObserver.createWithCleanup(self, log)
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPPasswordModifyRequest(
                    userIdentity="cn=another,ou=stuff,dc=example,dc=com",
                    newPasswd="hushhush",
                ),
                id=2,
            ).toWire()
        )
        self.assertEqual(
            observer[0]["log_text"],
            "User cn=thingie,ou=stuff,dc=example,dc=com "
            "tried to change password of "
            "b'cn=another,ou=stuff,dc=example,dc=com'",
        )
        self.assertListEqual(commits, [], "Server committed data.")
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPExtendedResponse(
                    resultCode=ldaperrors.LDAPInsufficientAccessRights.resultCode,
                    responseName=pureldap.LDAPPasswordModifyRequest.oid,
                ),
                id=2,
            ).toWire(),
        )
        self.assertSequenceEqual(
            self.thingie.get("userPassword", []),
            [userPassword],
        )

    def test_unknownRequest(self):
        # make server miss one of the handle_* attributes
        # without having to modify the LDAPServer class
        class MockServer(ldapserver.LDAPServer):
            handle_LDAPBindRequest = property()

        self.server.__class__ = MockServer
        self.server.dataReceived(
            pureldap.LDAPMessage(pureldap.LDAPBindRequest(), id=2).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPExtendedResponse(
                    resultCode=ldaperrors.LDAPProtocolError.resultCode,
                    responseName="1.3.6.1.4.1.1466.20036",
                    errorMessage="Unknown request",
                ),
                id=2,
            ).toWire(),
        )

    def test_control_unknown_critical(self):
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(),
                id=2,
                controls=[
                    ("42.42.42.42", True, None),
                ],
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.LDAPUnavailableCriticalExtension.resultCode,
                    errorMessage="Unknown control 42.42.42.42",
                ),
                id=2,
            ).toWire(),
        )

    def test_control_unknown_nonCritical(self):
        self.thingie["userPassword"] = [
            "{SSHA}yVLLj62rFf3kDAbzwEU0zYAVvbWrze8="
        ]  # "secret"
        self.server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(
                    dn="cn=thingie,ou=stuff,dc=example,dc=com", auth=b"secret"
                ),
                controls=[("42.42.42.42", False, None)],
                id=4,
            ).toWire()
        )
        self.assertEqual(
            self.server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=0, matchedDN="cn=thingie,ou=stuff,dc=example,dc=com"
                ),
                id=4,
            ).toWire(),
        )


class TestSchema(unittest.TestCase):
    def setUp(self):
        db = inmemory.ReadOnlyInMemoryLDAPEntry("", {})
        com = db.addChild(
            "dc=com",
            {
                "objectClass": ["dcObject"],
                "dc": ["com"],
            },
        )
        com.addChild(
            "dc=example",
            {
                "objectClass": ["dcObject"],
                "dc": ["example"],
                "subschemaSubentry": ["cn=schema"],
            },
        )
        db.addChild(
            "cn=schema",
            {
                "objectClass": ["TODO"],
                "cn": ["schema"],
                "attributeTypes": [
                    test_schema.AttributeType_KnownValues.knownValues[0][0]
                ],
                "objectClasses": [
                    test_schema.OBJECTCLASSES["organization"],
                    test_schema.OBJECTCLASSES["organizationalUnit"],
                ],
            },
        )

        class LDAPServerFactory(protocol.ServerFactory):
            protocol = ldapserver.LDAPServer

            def __init__(self, root):
                self.root = root

        components.registerAdapter(
            lambda x: x.root, LDAPServerFactory, interfaces.IConnectedLDAPEntry
        )
        serverFactory = LDAPServerFactory(db)

        self.client = ldapclient.LDAPClient()
        server = serverFactory.buildProtocol(
            address.IPv4Address("TCP", "localhost", "1024")
        )
        util.returnConnected(server, self.client)

    def testSimple(self):
        d = fetchschema.fetch(self.client, "dc=example,dc=com")
        (attributeTypes, objectClasses) = util.pumpingDeferredResult(d)

        self.failUnlessEqual(
            [to_bytes(x) for x in attributeTypes],
            [
                to_bytes(schema.AttributeTypeDescription(x))
                for x in [test_schema.AttributeType_KnownValues.knownValues[0][0]]
            ],
        )

        self.failUnlessEqual(
            [to_bytes(x) for x in objectClasses],
            [
                to_bytes(schema.ObjectClassDescription(x))
                for x in [
                    test_schema.OBJECTCLASSES["organization"],
                    test_schema.OBJECTCLASSES["organizationalUnit"],
                ]
            ],
        )

    testSimple.todo = "Not supported yet."



"""A Pure-Python Twisted library for LDAP"""
__version__ = "20.1.1"

__title__ = "ldaptor"
__description__ = "A Pure-Python Twisted library for LDAP"
__uri__ = "https://github.com/twisted/ldaptor"

__license__ = "MIT"
__author__ = "The ldaptor developers"
__copyright__ = "Copyright (c) 2002-2019 {}".format(__author__)



"""
    Test cases for ldaptor.encoder module
"""

from twisted.trial import unittest


import ldaptor._encoder


class WireableObject:
    """
    Object with bytes representation as a constant toWire value
    """

    def toWire(self):
        return b"wire"


class TextObject(ldaptor._encoder.TextStrAlias):
    """
    Object with human readable representation as a constant getText value
    """

    def getText(self):
        return "text"


class EncoderTests(unittest.TestCase):
    def test_wireable_object(self):
        """
        to_bytes function use object`s toWire method
        to get its bytes representation if it has one
        """
        obj = WireableObject()
        self.assertEqual(ldaptor._encoder.to_bytes(obj), b"wire")

    def test_unicode_object(self):
        """
        unicode string is encoded to utf-8 if passed
        to to_bytes function
        """
        obj = "unicode"
        self.assertEqual(ldaptor._encoder.to_bytes(obj), b"unicode")

    def test_bytes_object(self):
        """
        byte string is returned without changes
        if passed to to_bytes function
        """
        obj = b"bytes"
        self.assertEqual(ldaptor._encoder.to_bytes(obj), b"bytes")

    def test_int_object(self):
        """
        integer is converted to a string representation, then encoded to bytes
        if passed to to_bytes function
        """
        obj = 42
        self.assertEqual(ldaptor._encoder.to_bytes(obj), b"42")


class WireStrAliasTests(unittest.TestCase):
    def test_toWire_not_implemented(self):
        """
        WireStrAlias.toWire is an abstract method and raises NotImplementedError
        """
        obj = ldaptor._encoder.WireStrAlias()
        self.assertRaises(NotImplementedError, obj.toWire)


class TextStrAliasTests(unittest.TestCase):
    def test_deprecation_warning(self):
        str(TextObject())
        msg = (
            "TextObject.__str__ method is deprecated and will not be used "
            "for getting human readable representation in the future "
            "releases, use TextObject.getText instead"
        )
        warnings = self.flushWarnings()
        self.assertEqual(len(warnings), 1)
        self.assertEqual(warnings[0]["category"], DeprecationWarning)
        self.assertEqual(warnings[0]["message"], msg)

    def test_getText_not_implemented(self):
        """
        TextStrAlias.getText is an abstract method and raises NotImplementedError
        """
        obj = ldaptor._encoder.TextStrAlias()
        self.assertRaises(NotImplementedError, obj.getText)



#! /usr/bin/env python

"""
Example of using LDAPServer with the database.
Employees are stored in the database table and
retrieved on LDAPServerFactory initialization.
LDAP tree can be rebuilt by calling reload_tree method
if the list of users in the database has changed.
SQLAlchemy package is required to run this example.
"""

import sys

from ldaptor.inmemory import ReadOnlyInMemoryLDAPEntry
from ldaptor.interfaces import IConnectedLDAPEntry
from ldaptor.protocols.ldap.ldapserver import LDAPServer

from sqlalchemy import Column, Integer, String, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session

from twisted.internet import reactor
from twisted.internet.protocol import ServerFactory
from twisted.python.components import registerAdapter
from twisted.python import log

Base = declarative_base()


class LDAPServerFactory(ServerFactory):
    protocol = LDAPServer

    def __init__(self, db_engine):
        self.db_engine = db_engine
        self.tree = None
        self.reload_tree()

    def reload_tree(self):
        """
        Building LDAP tree.
        Call this method if you need to reload data from the database.
        """
        com_tree = ReadOnlyInMemoryLDAPEntry("dc=com")
        example_tree = com_tree.addChild("dc=example", {})
        users_tree = example_tree.addChild("ou=users", {})

        db_session = Session(self.db_engine)

        for employee in db_session.query(Employee):
            users_tree.addChild(
                "uid={}".format(employee.uid),
                {
                    "uid": [employee.uid],
                    "givenName": [employee.first_name],
                    "sn": [employee.last_name],
                    "email": [employee.email],
                },
            )

        db_session.close()

        self.tree = com_tree


class Employee(Base):
    __tablename__ = "employee"

    id = Column(Integer, primary_key=True)
    uid = Column(String(255), nullable=False)
    first_name = Column(String(255), nullable=False)
    last_name = Column(String(255), nullable=False)
    email = Column(String(255), nullable=False)


def create_db():
    """Creating a database with a table of employees and a couple of rows"""
    db_engine = create_engine("sqlite://")
    Base.metadata.bind = db_engine
    Employee.__table__.create()

    db_session = Session(db_engine)

    employee1 = Employee()
    employee1.uid = "f.example"
    employee1.first_name = "First"
    employee1.last_name = "Example"
    employee1.email = "first@example.com"
    db_session.add(employee1)

    employee2 = Employee()
    employee2.uid = "s.example"
    employee2.first_name = "Second"
    employee2.last_name = "Example"
    employee2.email = "second@example.com"
    db_session.add(employee2)

    db_session.commit()
    db_session.close()

    return db_engine


if __name__ == "__main__":
    engine = create_db()

    log.startLogging(sys.stderr)

    registerAdapter(lambda x: x.tree, LDAPServerFactory, IConnectedLDAPEntry)
    factory = LDAPServerFactory(engine)
    reactor.listenTCP(8080, factory)
    reactor.run()



"""
Test cases for ldaptor.protocols.ldap.fetchschema module.
"""

from twisted.trial import unittest
from ldaptor.protocols.ldap import fetchschema
from ldaptor import schema
from ldaptor.protocols import pureldap
from ldaptor.testutil import LDAPClientTestDriver
from ldaptor._encoder import to_bytes


class OnWire(unittest.TestCase):
    cn = """( 2.5.4.3 NAME ( 'cn' 'commonName' ) DESC 'RFC2256: common name(s) for which the entity is known by' SUP name )"""
    dcObject = """( 1.3.6.1.4.1.1466.344 NAME 'dcObject' DESC 'RFC2247: domain component object' SUP top AUXILIARY MUST dc )"""

    def testSimple(self):
        client = LDAPClientTestDriver(
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="",
                    attributes=(
                        ("subschemaSubentry", ["cn=Subschema"]),
                        ("bar", ["b", "c"]),
                    ),
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
            [
                pureldap.LDAPSearchResultEntry(
                    objectName="cn=Subschema",
                    attributes=(
                        ("attributeTypes", [self.cn]),
                        ("objectClasses", [self.dcObject]),
                    ),
                ),
                pureldap.LDAPSearchResultDone(
                    resultCode=0, matchedDN="", errorMessage=""
                ),
            ],
        )

        d = fetchschema.fetch(client, "dc=example,dc=com")
        d.addCallback(self._cb_testSimple, client)
        return d

    def _cb_testSimple(self, val, client):
        client.assertSent(
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                scope=pureldap.LDAP_SCOPE_baseObject,
                derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                sizeLimit=1,
                timeLimit=0,
                typesOnly=0,
                filter=pureldap.LDAPFilter_present("objectClass"),
                attributes=["subschemaSubentry"],
            ),
            pureldap.LDAPSearchRequest(
                baseObject="cn=Subschema",
                scope=pureldap.LDAP_SCOPE_baseObject,
                derefAliases=pureldap.LDAP_DEREF_neverDerefAliases,
                sizeLimit=1,
                timeLimit=0,
                typesOnly=0,
                filter=pureldap.LDAPFilter_present("objectClass"),
                attributes=["attributeTypes", "objectClasses"],
            ),
        )
        self.failUnlessEqual(len(val), 2)

        self.failUnlessEqual(
            [to_bytes(x) for x in val[0]],
            [to_bytes(schema.AttributeTypeDescription(self.cn))],
        )
        self.failUnlessEqual(
            [to_bytes(x) for x in val[1]],
            [to_bytes(schema.ObjectClassDescription(self.dcObject))],
        )



import os
from zope.interface import Interface, implements
from twisted.internet import reactor, defer
from nevow import rend, appserver, inevow, stan, loaders, flat
from formless import annotate, webform, iformless

from ldaptor.protocols.ldap import (
    ldapclient,
    ldapsyntax,
    ldapconnector,
    distinguishedname,
)
from ldaptor import ldapfilter
from ldaptor.protocols import pureldap


class ILDAPConfig(Interface):
    """Addressbook configuration retrieval."""

    def getBaseDN(self):
        """Get the LDAP base DN, as a DistinguishedName."""

    def getServiceLocationOverrides(self):
        """
        Get the LDAP service location overrides, as a mapping of
        DistinguishedName to (host, port) tuples.
        """


class LDAPConfig:
    implements(ILDAPConfig)

    def __init__(self, baseDN, serviceLocationOverrides=None):
        self.baseDN = distinguishedname.DistinguishedName(baseDN)
        self.serviceLocationOverrides = {}
        if serviceLocationOverrides is not None:
            for k, v in serviceLocationOverrides.items():
                dn = distinguishedname.DistinguishedName(k)
                self.serviceLocationOverrides[dn] = v

    def getBaseDN(self):
        return self.baseDN

    def getServiceLocationOverrides(self):
        return self.serviceLocationOverrides


class IAddressBookSearch(annotate.TypedInterface):
    def search(
        sn=annotate.String(label="Last name"),
        givenName=annotate.String(label="First name"),
        telephoneNumber=annotate.String(),
        description=annotate.String(),
    ):
        pass

    search = annotate.autocallable(search)


class CurrentSearch:
    implements(IAddressBookSearch, inevow.IContainer)
    data = {}

    def _getSearchFilter(self):
        filters = []
        for attr, value in self.data.items():
            if value is not None:
                f = ldapfilter.parseMaybeSubstring(attr, value)
                filters.append(f)
        if not filters:
            return None
        searchFilter = pureldap.LDAPFilter_and(
            [
                pureldap.LDAPFilter_equalityMatch(
                    attributeDesc=pureldap.LDAPAttributeDescription("objectClass"),
                    assertionValue=pureldap.LDAPAssertionValue("addressbookPerson"),
                )
            ]
            + filters
        )
        return searchFilter

    def search(self, **kw):
        for k, v in kw.items():
            if v is None:
                del kw[k]
        self.data = kw
        return self

    def __nonzero__(self):
        return bool(self.data)

    def __iter__(self):
        if self.data is None:
            return
        yield from self.data.items()

    def child(self, context, name):
        if name == "searchFilter":
            return self._getSearchFilter()
        if name != "results":
            return None
        config = context.locate(ILDAPConfig)

        c = ldapconnector.LDAPClientCreator(reactor, ldapclient.LDAPClient)
        d = c.connectAnonymously(
            config.getBaseDN(), config.getServiceLocationOverrides()
        )

        def _search(proto, base, searchFilter):
            baseEntry = ldapsyntax.LDAPEntry(client=proto, dn=base)
            d = baseEntry.search(filterObject=searchFilter)
            return d

        d.addCallback(_search, config.getBaseDN(), self._getSearchFilter())
        return d


def LDAPFilterSerializer(original, context):
    return original.asText()


# TODO need to make this pretty some day.
for c in [
    pureldap.LDAPFilter_and,
    pureldap.LDAPFilter_or,
    pureldap.LDAPFilter_not,
    pureldap.LDAPFilter_substrings,
    pureldap.LDAPFilter_equalityMatch,
    pureldap.LDAPFilter_greaterOrEqual,
    pureldap.LDAPFilter_lessOrEqual,
    pureldap.LDAPFilter_approxMatch,
    pureldap.LDAPFilter_present,
    pureldap.LDAPFilter_extensibleMatch,
]:
    flat.registerFlattener(LDAPFilterSerializer, c)


class AddressBookResource(rend.Page):
    docFactory = loaders.xmlfile(
        "searchform.xhtml", templateDir=os.path.split(os.path.abspath(__file__))[0]
    )

    def configurable_(self, context):
        try:
            i = context.locate(inevow.IHand)
        except KeyError:
            i = CurrentSearch()
        return i

    def data_search(self, context, data):
        d = defer.maybeDeferred(self.locateConfigurable, context, "")

        def cb(configurable):
            return configurable.original

        d.addCallback(cb)
        return d

    def child_form_css(self, request):
        return webform.defaultCSS

    def render_input(self, context, data):
        formDefaults = context.locate(iformless.IFormDefaults)
        methodDefaults = formDefaults.getAllDefaults("search")
        conf = self.configurable_(context)
        for k, v in conf:
            methodDefaults[k] = v
        return webform.renderForms()

    def render_haveSearch(self, context, data):
        r = context.tag.allPatterns(str(bool(data)))
        return context.tag.clear()[r]

    def render_searchFilter(self, context, data):
        return data.asText()

    def render_iterateMapping(self, context, data):
        headers = context.tag.allPatterns("header")
        keyPattern = context.patternGenerator("key")
        valuePattern = context.patternGenerator("value")
        divider = context.patternGenerator("divider", default=stan.invisible)
        content = [
            (keyPattern(data=key), valuePattern(data=value), divider())
            for key, value in data.items()
        ]
        if not content:
            content = context.tag.allPatterns("empty")
        else:
            # No divider after the last thing.
            content[-1] = content[-1][:-1]
        footers = context.tag.allPatterns("footer")

        return context.tag.clear()[headers, content, footers]


def getSite(config):
    form = AddressBookResource()
    form.remember(config, ILDAPConfig)
    site = appserver.NevowSite(form)
    return site



#!/usr/bin/env python3
# Copyright (C) 2001 Tommi Virtanen
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of version 2.1 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from ldaptor.protocols.ldap import (
    ldapclient,
    ldapsyntax,
    ldapconnector,
    distinguishedname,
)
from twisted.internet import defer, reactor
from twisted.python import log
import sys

exitStatus = 0


def error(fail):
    print("fail:", fail.getErrorMessage(), file=sys.stderr)
    global exitStatus
    exitStatus = 1


def _handle_entry(entry, connection, search):
    sys.stdout.write("# connection %d, search %d\n%s" % (connection, search, entry))


def _search(proto, base, connection, numOfSearches):
    l = []
    baseEntry = ldapsyntax.LDAPEntry(client=proto, dn=base)
    for search in range(0, numOfSearches):
        d = baseEntry.search(callback=lambda x: _handle_entry(x, connection, search))
        d.addErrback(error)
        l.append(d)

    dl = defer.DeferredList(l)
    return dl


def main(base, serviceLocationOverrides, numOfConnections=3, numOfSearches=3):
    log.startLogging(sys.stderr, setStdout=0)
    l = []
    for connection in range(0, numOfConnections):
        c = ldapconnector.LDAPClientCreator(reactor, ldapclient.LDAPClient)
        d = c.connectAnonymously(base, serviceLocationOverrides)

        d.addCallback(_search, base, connection, numOfSearches)
        d.addErrback(error)
        l.append(d)
    dl = defer.DeferredList(l)
    dl.addBoth(lambda dummy: reactor.stop())
    reactor.run()
    sys.exit(exitStatus)


if __name__ == "__main__":
    base = "dc=example,dc=com"
    main(
        base=distinguishedname.DistinguishedName(base),
        serviceLocationOverrides={base: ("localhost", None)},
        numOfConnections=5,
        numOfSearches=10,
    )



from ldaptor.protocols.ldap.autofill import ObjectMissingObjectClassException


class Autofill_samba:  # TODO baseclass
    def start(self, ldapObject):
        assert "objectClass" in ldapObject
        if "sambaAccount" not in ldapObject["objectClass"]:
            raise ObjectMissingObjectClassException(ldapObject)

        assert "acctFlags" not in ldapObject
        ldapObject["acctFlags"] = ["[UX         ]"]
        assert "pwdLastSet" not in ldapObject
        ldapObject["pwdLastSet"] = ["0"]
        assert "logonTime" not in ldapObject
        ldapObject["logonTime"] = ["0"]
        assert "logoffTime" not in ldapObject
        ldapObject["logoffTime"] = ["0"]
        assert "pwdCanChange" not in ldapObject
        ldapObject["pwdCanChange"] = ["0"]
        assert "pwdMustChange" not in ldapObject
        ldapObject["pwdMustChange"] = ["0"]

    def notify(self, ldapObject, attributeType):

        # rid=2*uid+1000
        if attributeType == "uidNumber":
            assert "uidNumber" in ldapObject
            assert len(ldapObject["uidNumber"]) == 1
            for uidNumber in ldapObject["uidNumber"]:
                uidNumber = int(uidNumber)
                rid = uidNumber * 2 + 1000
                ldapObject["rid"] = [str(rid)]
                return

        # primaryGroupID=2*gid+1001
        if attributeType == "gidNumber":
            assert "gidNumber" in ldapObject
            assert len(ldapObject["gidNumber"]) == 1
            for gidNumber in ldapObject["gidNumber"]:
                gidNumber = int(gidNumber)
                primaryGroupID = gidNumber * 2 + 1001
                ldapObject["primaryGroupID"] = [str(primaryGroupID)]
                return



"""
Test cases for ldaptor.protocols.ldap.svcbindproxy module.
"""

from twisted.trial import unittest
from twisted.internet import reactor
from ldaptor.protocols.ldap import svcbindproxy, ldaperrors
from ldaptor.protocols import pureldap, pureber
from ldaptor import ldapfilter, testutil


class ServiceBindingProxy(unittest.TestCase):
    berdecoder = pureldap.LDAPBERDecoderContext_TopLevel(
        inherit=pureldap.LDAPBERDecoderContext_LDAPMessage(
            fallback=pureldap.LDAPBERDecoderContext(
                fallback=pureber.BERDecoderContext()
            ),
            inherit=pureldap.LDAPBERDecoderContext(
                fallback=pureber.BERDecoderContext()
            ),
        )
    )

    def createServer(self, services, fallback=None, responses=[]):
        server = testutil.createServer(
            lambda config: svcbindproxy.ServiceBindingProxy(
                config=config,
                services=services,
                fallback=fallback,
            ),
            baseDN="dc=example,dc=com",
            *responses,
        )
        server.now = "20050213140302Z"
        server.timestamp = lambda: server.now
        return server

    def test_bind_noMatchingServicesFound_noFallback(self):
        server = self.createServer(
            services=[
                "svc1",
                "svc2",
                "svc3",
            ],
            fallback=False,
            responses=[
                [pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode)],
                [pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode)],
                [pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode)],
            ],
        )
        server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(dn="cn=jack,dc=example,dc=com", auth="s3krit"),
                id=4,
            ).toWire()
        )
        reactor.iterate()  # TODO
        client = server.client

        client.assertSent(
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc1)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc2)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc3)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
        )
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.LDAPInvalidCredentials.resultCode
                ),
                id=4,
            ).toWire(),
        )

    def test_bind_noMatchingServicesFound_fallback_success(self):
        server = self.createServer(
            services=[
                "svc1",
                "svc2",
                "svc3",
            ],
            fallback=True,
            responses=[
                [pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode)],
                [pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode)],
                [pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode)],
                [pureldap.LDAPBindResponse(resultCode=ldaperrors.Success.resultCode)],
            ],
        )
        server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(dn="cn=jack,dc=example,dc=com", auth="s3krit"),
                id=4,
            ).toWire()
        )
        reactor.iterate()  # TODO
        client = server.client

        client.assertSent(
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc1)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc2)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc3)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
            pureldap.LDAPBindRequest(dn="cn=jack,dc=example,dc=com", auth="s3krit"),
        )
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(resultCode=ldaperrors.Success.resultCode),
                id=4,
            ).toWire(),
        )

    def test_bind_noMatchingServicesFound_fallback_badAuth(self):
        server = self.createServer(
            services=[
                "svc1",
                "svc2",
                "svc3",
            ],
            fallback=True,
            responses=[
                [pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode)],
                [pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode)],
                [pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode)],
                [
                    pureldap.LDAPBindResponse(
                        resultCode=ldaperrors.LDAPInvalidCredentials.resultCode
                    ),
                ],
            ],
        )
        server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(
                    dn="cn=jack,dc=example,dc=com", auth="wrong-s3krit"
                ),
                id=4,
            ).toWire()
        )
        reactor.iterate()  # TODO
        client = server.client

        client.assertSent(
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc1)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc2)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc3)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
            pureldap.LDAPBindRequest(
                dn="cn=jack,dc=example,dc=com", auth="wrong-s3krit"
            ),
        )
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.LDAPInvalidCredentials.resultCode
                ),
                id=4,
            ).toWire(),
        )

    def test_bind_match_success(self):
        server = self.createServer(
            services=[
                "svc1",
                "svc2",
                "svc3",
            ],
            fallback=True,
            responses=[
                # svc1
                [
                    pureldap.LDAPSearchResultEntry(
                        r"cn=svc1+owner=cn\=jack\,dc\=example\,dc\=com,dc=example,dc=com",
                        attributes=[],
                    ),
                    pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode),
                ],
                [pureldap.LDAPBindResponse(resultCode=ldaperrors.Success.resultCode)],
            ],
        )

        server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(dn="cn=jack,dc=example,dc=com", auth="secret"),
                id=4,
            ).toWire()
        )
        reactor.iterate()  # TODO
        client = server.client

        client.assertSent(
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc1)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
            pureldap.LDAPBindRequest(
                dn=r"cn=svc1+owner=cn\=jack\,dc\=example\,dc\=com,dc=example,dc=com",
                auth="secret",
            ),
        )
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.Success.resultCode,
                    matchedDN="cn=jack,dc=example,dc=com",
                ),
                id=4,
            ).toWire(),
        )

    def test_bind_match_success_later(self):
        server = self.createServer(
            services=[
                "svc1",
                "svc2",
                "svc3",
            ],
            fallback=True,
            responses=[
                # svc1
                [
                    pureldap.LDAPSearchResultEntry(
                        r"cn=svc1+owner=cn\=jack\,dc\=example\,dc\=com,dc=example,dc=com",
                        attributes=[],
                    ),
                    pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode),
                ],
                [
                    pureldap.LDAPBindResponse(
                        resultCode=ldaperrors.LDAPInvalidCredentials.resultCode
                    )
                ],
                # svc2
                [pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode)],
                # svc3
                [
                    pureldap.LDAPSearchResultEntry(
                        r"cn=svc3+owner=cn\=jack\,dc\=example\,dc\=com,dc=example,dc=com",
                        attributes=[],
                    ),
                    pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode),
                ],
                [pureldap.LDAPBindResponse(resultCode=ldaperrors.Success.resultCode)],
            ],
        )

        server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(dn="cn=jack,dc=example,dc=com", auth="secret"),
                id=4,
            ).toWire()
        )
        reactor.iterate()  # TODO
        client = server.client

        client.assertSent(
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc1)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
            pureldap.LDAPBindRequest(
                dn=r"cn=svc1+owner=cn\=jack\,dc\=example\,dc\=com,dc=example,dc=com",
                auth="secret",
            ),
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc2)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc3)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
            pureldap.LDAPBindRequest(
                dn=r"cn=svc3+owner=cn\=jack\,dc\=example\,dc\=com,dc=example,dc=com",
                auth="secret",
            ),
        )
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.Success.resultCode,
                    matchedDN="cn=jack,dc=example,dc=com",
                ),
                id=4,
            ).toWire(),
        )

    def test_bind_match_badAuth(self):
        server = self.createServer(
            services=[
                "svc1",
                "svc2",
                "svc3",
            ],
            fallback=True,
            responses=[
                # svc1
                [
                    pureldap.LDAPSearchResultEntry(
                        r"cn=svc1+owner=cn\=jack\,dc\=example\,dc\=com,dc=example,dc=com",
                        attributes=[],
                    ),
                    pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode),
                ],
                [
                    pureldap.LDAPBindResponse(
                        resultCode=ldaperrors.LDAPInvalidCredentials.resultCode
                    )
                ],
                # svc2
                [pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode)],
                # svc3
                [
                    pureldap.LDAPSearchResultEntry(
                        r"cn=svc3+owner=cn\=jack\,dc\=example\,dc\=com,dc=example,dc=com",
                        attributes=[],
                    ),
                    pureldap.LDAPSearchResultDone(ldaperrors.Success.resultCode),
                ],
                [
                    pureldap.LDAPBindResponse(
                        resultCode=ldaperrors.LDAPInvalidCredentials.resultCode
                    )
                ],
                [
                    pureldap.LDAPBindResponse(
                        resultCode=ldaperrors.LDAPInvalidCredentials.resultCode
                    )
                ],
            ],
        )

        server.dataReceived(
            pureldap.LDAPMessage(
                pureldap.LDAPBindRequest(
                    dn="cn=jack,dc=example,dc=com", auth="wrong-s3krit"
                ),
                id=4,
            ).toWire()
        )
        reactor.iterate()  # TODO
        client = server.client

        client.assertSent(
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc1)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
            pureldap.LDAPBindRequest(
                dn=r"cn=svc1+owner=cn\=jack\,dc\=example\,dc\=com,dc=example,dc=com",
                auth="wrong-s3krit",
            ),
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc2)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
            pureldap.LDAPSearchRequest(
                baseObject="dc=example,dc=com",
                derefAliases=0,
                sizeLimit=0,
                timeLimit=0,
                typesOnly=0,
                filter=ldapfilter.parseFilter(
                    "(&"
                    + "(objectClass=serviceSecurityObject)"
                    + "(owner=cn=jack,dc=example,dc=com)"
                    + "(cn=svc3)"
                    + ("(|(!(validFrom=*))(validFrom<=%s))" % server.now)
                    + ("(|(!(validUntil=*))(validUntil>=%s))" % server.now)
                    + ")"
                ),
                attributes=("1.1",),
            ),
            pureldap.LDAPBindRequest(
                dn=r"cn=svc3+owner=cn\=jack\,dc\=example\,dc\=com,dc=example,dc=com",
                auth="wrong-s3krit",
            ),
            pureldap.LDAPBindRequest(
                version=3, dn="cn=jack,dc=example,dc=com", auth="wrong-s3krit"
            ),
        )
        self.assertEqual(
            server.transport.value(),
            pureldap.LDAPMessage(
                pureldap.LDAPBindResponse(
                    resultCode=ldaperrors.LDAPInvalidCredentials.resultCode
                ),
                id=4,
            ).toWire(),
        )



#!/usr/bin/env python3
from twisted.internet import reactor, defer

from ldaptor.protocols.ldap import (
    ldapclient,
    ldapsyntax,
    ldapconnector,
    distinguishedname,
)
from ldaptor import ldapfilter


def search(config):
    c = ldapconnector.LDAPClientCreator(reactor, ldapclient.LDAPClient)
    d = c.connectAnonymously(config["base"], config["serviceLocationOverrides"])

    def _doSearch(proto, config):
        searchFilter = ldapfilter.parseFilter("(gn=j*)")
        baseEntry = ldapsyntax.LDAPEntry(client=proto, dn=config["base"])
        d = baseEntry.search(filterObject=searchFilter)
        return d

    d.addCallback(_doSearch, config)
    return d


def main():
    import sys
    from twisted.python import log

    log.startLogging(sys.stderr, setStdout=0)

    config = {
        "base": distinguishedname.DistinguishedName("ou=People,dc=example,dc=com"),
        "serviceLocationOverrides": {
            distinguishedname.DistinguishedName("dc=example,dc=com"): (
                "localhost",
                10389,
            ),
        },
    }

    d = search(config)

    def _show(results):
        for item in results:
            print(item)

    d.addCallback(_show)
    d.addErrback(defer.logError)
    d.addBoth(lambda _: reactor.stop())
    reactor.run()


if __name__ == "__main__":
    main()
